"""
The equilibrium module defines routines for interacting with
calculated phase equilibria.
"""
from __future__ import print_function
import pycalphad.variables as v
from pycalphad.core.utils import unpack_kwarg
from pycalphad.core.utils import sizeof_fmt
from pycalphad.core.utils import unpack_condition, unpack_phases
from pycalphad import calculate, Model
from pycalphad.constraints import mole_fraction
from pycalphad.core.lower_convex_hull import lower_convex_hull
from pycalphad.core.autograd_utils import build_functions
from pycalphad.core.constants import MIN_SITE_FRACTION, COMP_DIFFERENCE_TOL
from sympy import Add, Symbol
from tqdm import tqdm as progressbar
from xarray import Dataset, DataArray
import numpy as np
import scipy.spatial
from collections import defaultdict, namedtuple, OrderedDict
import itertools
import copy
from datetime import datetime
import sys

# Maximum number of global search iterations
MAX_SEARCH_ITERATIONS = 30
# Maximum number of multi-phase solver iterations
MAX_SOLVE_ITERATIONS = 30
# Minimum step norm size to stop solver
MIN_SOLVE_STEP_NORM = 1e-6
# Minimum step multiplier to stop line search
MIN_SOLVE_ALPHA = 1e-6
# Minimum energy (J/mol-atom) difference between iterations before stopping solver
MIN_SOLVE_ENERGY_PROGRESS = 1e-6
# Maximum number of backtracking iterations
MAX_BACKTRACKING = 5
# Maximum number of Newton steps to take
MAX_NEWTON_ITERATIONS = 1
# If the max of the potential difference between iterations is less than
# MIN_SEARCH_PROGRESS J/mol-atom, stop the global search
MIN_SEARCH_PROGRESS = np.inf
# Minimum norm of a Newton direction before it's "zero"
MIN_DIRECTION_NORM = 1e-12
# initial value of 'alpha' in Newton-Raphson procedure
INITIAL_STEP_SIZE = 1.

PhaseRecord = namedtuple('PhaseRecord', ['variables', 'grad', 'hess', 'plane_grad', 'plane_hess',
                                         'mass_obj', 'mass_grad', 'mass_hess'])

class EquilibriumError(Exception):
    "Exception related to calculation of equilibrium"
    pass


class ConditionError(EquilibriumError):
    "Exception related to equilibrium conditions"
    pass


def remove_degenerate_phases(properties, multi_index):
    """
    For each phase pair with composition difference below tolerance,
    eliminate phase with largest index.

    Parameters
    ----------
    properties : xarray.Dataset
        Equilibrium calculation data. This will be modified!
    multi_index : tuple
        Index into 'properties' of the condition set of interest.

    """
    phases = list(properties['Phase'].values[multi_index])
    # Are there already removed phases?
    if '' in phases:
        num_phases = phases.index('')
    else:
        num_phases = len(phases)
    phases = properties['Phase'].values[multi_index + np.index_exp[:num_phases]]
    # Group phases into multiple composition sets
    phase_indices = defaultdict(lambda: list())
    for phase_idx, name in enumerate(phases):
        phase_indices[name].append(phase_idx)
    # Compute pairwise distances between compositions of like phases
    for name, indices in phase_indices.items():
        if len(indices) == 1:
            # Phase is unique
            continue
        # The reason we don't do this based on Y fractions is because
        # of sublattice symmetry. It's very easy to detect a "miscibility gap" which is actually
        # symmetry equivalent, i.e., D([A, B] - [B, A]) > tol, but they are the same configuration.
        comp_matrix = properties['X'].values[multi_index + np.index_exp[indices]]
        comp_distances = scipy.spatial.distance.squareform(scipy.spatial.distance.pdist(comp_matrix, metric='chebyshev'))
        redundant_phases = set()
        redundant_phases |= {indices[0]}
        for i in range(len(indices)):
            for j in range(i, len(indices)):
                if i == j:
                    continue
                if comp_distances[i, j] < COMP_DIFFERENCE_TOL:
                    redundant_phases |= {indices[i], indices[j]}
        redundant_phases = sorted(redundant_phases)
        kept_phase = redundant_phases[0]
        removed_phases = redundant_phases[1:]
        # Their NP values will be added to redundant_phases[0]
        # and they will be nulled out
        for redundant in removed_phases:
            properties['NP'].values[multi_index + np.index_exp[kept_phase]] += \
                properties['NP'].values[multi_index + np.index_exp[redundant]]
            properties['Phase'].values[multi_index + np.index_exp[redundant]] = ''
    # Eliminate any 'fake points' that made it through the convex hull routine
    # These can show up from phases which aren't defined over all of composition space
    properties['NP'].values[np.nonzero(properties['Phase'].values == '_FAKE_')] = np.nan
    properties['Phase'].values[np.nonzero(properties['Phase'].values == '_FAKE_')] = ''
    # Rewrite properties to delete all the nulled out phase entries
    # Then put them at the end
    # That will let us rewrite 'phases' to have only the independent phases
    # And still preserve convenient indexing of 'properties' with phase_idx
    saved_indices = properties['Phase'].values[multi_index] != ''
    saved_indices = np.arange(len(saved_indices))[saved_indices]
    # TODO: Assumes N=1 always
    properties['NP'].values[multi_index + np.index_exp[:len(saved_indices)]] = \
        properties['NP'].values[multi_index + np.index_exp[saved_indices]] / \
        np.sum(properties['NP'].values[multi_index + np.index_exp[saved_indices]])
    properties['NP'].values[multi_index + np.index_exp[len(saved_indices):]] = np.nan
    properties['Phase'].values[multi_index + np.index_exp[:len(saved_indices)]] = \
        properties['Phase'].values[multi_index + np.index_exp[saved_indices]]
    properties['Phase'].values[multi_index + np.index_exp[len(saved_indices):]] = ''
    properties['X'].values[multi_index + np.index_exp[:len(saved_indices), :]] = \
        properties['X'].values[multi_index + np.index_exp[saved_indices, :]]
    properties['X'].values[multi_index + np.index_exp[len(saved_indices):, :]] = np.nan
    properties['Y'].values[multi_index + np.index_exp[:len(saved_indices), :]] = \
        properties['Y'].values[multi_index + np.index_exp[saved_indices, :]]
    properties['Y'].values[multi_index + np.index_exp[len(saved_indices):, :]] = np.nan


def _adjust_conditions(conds):
    "Adjust conditions values to be within the numerical limit of the solver."
    new_conds = OrderedDict()
    for key, value in sorted(conds.items(), key=str):
        if isinstance(key, v.Composition):
            new_conds[key] = [max(val, MIN_SITE_FRACTION*1000) for val in unpack_condition(value)]
        else:
            new_conds[key] = unpack_condition(value)
    return new_conds

def _compute_phase_dof(dbf, comps, phases):
    """
    Generate a list of the number of each phase's internal phase degrees of freedom.
    """
    phase_dof = []
    for name in phases:
        total = 0
        for idx in range(len(dbf.phases[name].sublattices)):
            active_in_subl = set(dbf.phases[name].constituents[idx]).intersection(comps)
            total += len(active_in_subl)
        phase_dof.append(total)
    return phase_dof

def _compute_constraints(dbf, comps, phases, cur_conds, site_fracs, phase_fracs, phase_records,
                         l_multipliers=None, chem_pots=None):
    """
    Compute the constraint vector and constraint Jacobian matrix.
    """
    num_sitefrac_bals = sum([len(dbf.phases[i].sublattices) for i in phases])
    num_mass_bals = len([i for i in cur_conds.keys() if i.startswith('X_')]) + 1
    indep_sum = np.sum([float(val) for i, val in cur_conds.items() if i.startswith('X_')])
    dependent_comp = set(comps) - set([i[2:] for i in cur_conds.keys() if i.startswith('X_')]) - {'VA'}
    dependent_comp = list(dependent_comp)[0]
    mole_fractions = {}
    num_constraints = num_sitefrac_bals + num_mass_bals
    num_vars = len(site_fracs) + len(phases)
    phase_dof = _compute_phase_dof(dbf, comps, phases)
    l_constraints = np.zeros(num_constraints, dtype=np.float)
    if l_multipliers is None:
        l_multipliers = np.zeros(num_constraints, dtype=np.float)
    if chem_pots is not None:
        # Insert chemical potentials as Lagrange multipliers of mass balance constraints
        l_multipliers[sum([len(dbf.phases[i].sublattices) for i in phases]):
        sum([len(dbf.phases[i].sublattices) for i in phases]) + num_mass_bals] = chem_pots
    # Convenience object for caller so it doesn't need to know about the constraint configuration
    chemical_potentials = l_multipliers[sum([len(dbf.phases[i].sublattices) for i in phases]):
        sum([len(dbf.phases[i].sublattices) for i in phases]) + num_mass_bals]

    constraint_jac = np.zeros((num_constraints, num_vars), dtype=np.float)
    # Ordering of constraints by row: sitefrac bal of each phase, then component mass balance
    # Ordering of constraints by column: site fractions of each phase, then phase fractions
    # First: Site fraction balance constraints
    var_idx = 0
    constraint_offset = 0
    for name in phases:
        for idx in range(len(dbf.phases[name].sublattices)):
            active_in_subl = set(dbf.phases[name].constituents[idx]).intersection(comps)
            constraint_jac[constraint_offset + idx,
            var_idx:var_idx + len(active_in_subl)] = 1
            # print('L_CONSTRAINTS[{}] = {}'.format(constraint_offset+idx, (sum(site_fracs[var_idx:var_idx + len(active_in_subl)]) - 1)))
            l_constraints[constraint_offset + idx] = \
                (sum(site_fracs[var_idx:var_idx + len(active_in_subl)]) - 1)
            var_idx += len(active_in_subl)
        constraint_offset += len(dbf.phases[name].sublattices)
    # Second: Mass balance of each component
    for comp in [c for c in comps if c != 'VA']:
        var_offset = 0
        phase_idx = 0
        for name, phase_frac in zip(phases, phase_fracs):
            if mole_fractions.get((name, comp), None) is None:
                mole_fractions[(name, comp)] = build_functions(mole_fraction(dbf.phases[name], comps, comp),
                                                               sorted(set(phase_records[name].variables) - {v.T, v.P},
                                                                      key=str))
            comp_obj, comp_grad, comp_hess = mole_fractions[(name, comp)]
            # current phase frac times the comp_grad
            constraint_jac[constraint_offset,
            var_offset:var_offset + phase_dof[phase_idx]] = \
                phase_frac * np.squeeze(comp_grad(*site_fracs[var_offset:var_offset + phase_dof[phase_idx]]))
            constraint_jac[constraint_offset, len(site_fracs) + phase_idx] += \
                np.squeeze(comp_obj(*site_fracs[var_offset:var_offset + phase_dof[phase_idx]]))
            l_constraints[constraint_offset] += \
                phase_frac * np.squeeze(comp_obj(*site_fracs[var_offset:var_offset + phase_dof[phase_idx]]))
            # print('L_CONSTRAINTS[{}] += {}'.format(constraint_offset, phase_frac * np.squeeze(comp_obj(*site_fracs[var_offset:var_offset+phase_dof[phase_idx]]))))
            var_offset += phase_dof[phase_idx]
            phase_idx += 1
        if comp != dependent_comp:
            l_constraints[constraint_offset] -= float(cur_conds['X_' + comp])
            # print('L_CONSTRAINTS[{}] -= {}'.format(constraint_offset, float(cur_conds['X_'+comp])))
        else:
            # TODO: Assuming N=1 (fixed for dependent component)
            l_constraints[constraint_offset] -= (1 - indep_sum)
            # print('L_CONSTRAINTS[{}] -= {}'.format(constraint_offset, (1-indep_sum)))
        #l_constraints[constraint_offset] *= -1
        # print('L_CONSTRAINTS[{}] *= -1'.format(constraint_offset))
        constraint_offset += 1
    return l_constraints, constraint_jac, l_multipliers, chemical_potentials, mole_fractions

def _compute_multiphase_objective(dbf, comps, phases, cur_conds, site_fracs, phase_fracs, callable_dict):
    result = 0
    phase_dof = _compute_phase_dof(dbf, comps, phases)
    var_offset = 0
    for phase_idx, (name, phase_frac) in enumerate(zip(phases, phase_fracs)):
        obj = callable_dict[name]
        obj_res = obj(*itertools.chain([cur_conds['P'], cur_conds['T']],
                                       site_fracs[var_offset:var_offset + phase_dof[phase_idx]])
                      )
        result += phase_frac * obj_res
        var_offset += phase_dof[phase_idx]
    return result

def _build_multiphase_gradient(dbf, comps, phases, cur_conds, site_fracs, phase_fracs,
                               l_constraints, constraint_jac, l_multipliers, callable_dict, phase_records):
    var_offset = 0
    phase_idx = 0
    phase_dof = _compute_phase_dof(dbf, comps, phases)
    num_vars = len(site_fracs) + len(phases)
    num_constraints = len(l_constraints)
    gradient_term = np.zeros((num_vars + num_constraints), dtype=np.float)
    for name, phase_frac in zip(phases, phase_fracs):
        obj = callable_dict[name]
        grad = phase_records[name].grad
        obj_res = obj(*itertools.chain([cur_conds['P'], cur_conds['T']],
                                       site_fracs[var_offset:var_offset + phase_dof[phase_idx]])
                      )
        grad_res = grad(*itertools.chain([cur_conds['P'], cur_conds['T']],
                                         site_fracs[var_offset:var_offset + phase_dof[phase_idx]])
                        )
        gradient_term[var_offset:var_offset + phase_dof[phase_idx]] = \
            -phase_frac * np.squeeze(grad_res)[2:]  # Remove P,T grad part
        gradient_term[len(site_fracs) + phase_idx] = -obj_res
        var_offset += phase_dof[phase_idx]
        phase_idx += 1
    gradient_term[:num_vars] += np.dot(constraint_jac.T, l_multipliers)
    gradient_term[num_vars:] = -l_constraints
    return gradient_term

def _build_multiphase_system(dbf, comps, phases, cur_conds, site_fracs, phase_fracs,
                             l_constraints, constraint_jac, l_multipliers, callable_dict, phase_records):
    # Now build objective Hessian and gradient terms
    var_offset = 0
    phase_idx = 0
    phase_dof = _compute_phase_dof(dbf, comps, phases)
    num_vars = len(site_fracs) + len(phases)
    num_constraints = len(l_constraints)
    l_hessian = np.zeros((num_vars + num_constraints, num_vars + num_constraints), dtype=np.float)
    gradient_term = np.zeros((num_vars + num_constraints), dtype=np.float)
    for name, phase_frac in zip(phases, phase_fracs):
        obj = callable_dict[name]
        hess = phase_records[name].hess
        grad = phase_records[name].grad
        obj_res = obj(*itertools.chain([cur_conds['P'], cur_conds['T']],
                                       site_fracs[var_offset:var_offset + phase_dof[phase_idx]])
                      )
        grad_res = grad(*itertools.chain([cur_conds['P'], cur_conds['T']],
                                         site_fracs[var_offset:var_offset + phase_dof[phase_idx]])
                        )
        gradient_term[var_offset:var_offset + phase_dof[phase_idx]] = \
            -phase_frac * np.squeeze(grad_res)[2:]  # Remove P,T grad part
        gradient_term[len(site_fracs) + phase_idx] = -obj_res
        l_hessian[var_offset:var_offset + phase_dof[phase_idx],
        var_offset:var_offset + phase_dof[phase_idx]] = \
            phase_frac * np.squeeze(hess(*itertools.chain([cur_conds['P'], cur_conds['T']],
                                                          site_fracs[var_offset:var_offset + phase_dof[phase_idx]])
                                         ))[2:, 2:]  # Remove P,T hessian part
        # Phase fraction / site fraction cross derivative
        l_hessian[len(site_fracs) + phase_idx, var_offset:var_offset + phase_dof[phase_idx]] = \
            l_hessian[var_offset:var_offset + phase_dof[phase_idx], len(site_fracs) + phase_idx] = \
            (gradient_term[var_offset:var_offset + phase_dof[phase_idx]] / -phase_frac)
        var_offset += phase_dof[phase_idx]
        phase_idx += 1
    # TODO: N=1 fixed by default
    # l_constraints[constraint_offset] -= 1
    # print('L_CONSTRAINTS', l_constraints.astype(np.longfloat))
    # print('CONSTRAINT_JAC', constraint_jac)
    l_hessian[:num_vars, num_vars:] = -constraint_jac.T
    l_hessian[num_vars:, :num_vars] = constraint_jac
    # print('PURE GRADIENT', gradient_term)
    gradient_term[:num_vars] += np.dot(constraint_jac.T, l_multipliers)
    gradient_term[num_vars:] = -l_constraints
    return l_hessian, gradient_term

def _eqcalculate(dbf, comps, phases, conditions, output, data=None, per_phase=False, **kwargs):
    """
    WARNING: API/calling convention not finalized.
    Compute the *equilibrium value* of a property.
    This function differs from `calculate` in that it computes
    thermodynamic equilibrium instead of randomly sampling the
    internal degrees of freedom of a phase.
    Because of that, it's slower than `calculate`.
    This plugs in the equilibrium phase and site fractions
    to compute a thermodynamic property defined in a Model.

    Parameters
    ----------
    dbf : Database
        Thermodynamic database containing the relevant parameters.
    comps : list
        Names of components to consider in the calculation.
    phases : list or dict
        Names of phases to consider in the calculation.
    conditions : dict or (list of dict)
        StateVariables and their corresponding value.
    output : str
        Equilibrium model property (e.g., CPM, HM, etc.) to compute.
        This must be defined as an attribute in the Model class of each phase.
    data : Dataset, optional
        Previous result of call to `equilibrium`.
        Should contain the equilibrium configurations at the conditions of interest.
        If the databases are not the same as in the original calculation,
        the results may be meaningless. If None, `equilibrium` will be called.
        Specifying this keyword argument can save the user some time if several properties
        need to be calculated in succession.
    per_phase : bool, optional
        If True, compute and return the property for each phase present.
        If False, return the total system value, weighted by the phase fractions.
    kwargs
        Passed to `calculate`.

    Returns
    -------
    Dataset of property as a function of equilibrium conditions
    """
    if data is None:
        data = equilibrium(dbf, comps, phases, conditions)
    active_phases = unpack_phases(phases) or sorted(dbf.phases.keys())
    conds = _adjust_conditions(conditions)
    indep_vars = ['P', 'T']
    # TODO: Rewrite this to use the coord dict from 'data'
    str_conds = OrderedDict((str(key), value) for key, value in conds.items())
    indep_vals = list([float(x) for x in np.atleast_1d(val)]
                      for key, val in str_conds.items() if key in indep_vars)
    coord_dict = str_conds.copy()
    components = [x for x in sorted(comps) if not x.startswith('VA')]
    coord_dict['vertex'] = np.arange(len(components))
    grid_shape = np.meshgrid(*coord_dict.values(),
                             indexing='ij', sparse=False)[0].shape
    prop_shape = grid_shape
    prop_dims = list(str_conds.keys()) + ['vertex']

    result = Dataset({output: (prop_dims, np.full(prop_shape, np.nan))}, coords=coord_dict)
    # For each phase select all conditions where that phase exists
    # Perform the appropriate calculation and then write the result back
    for phase in active_phases:
        dof = sum([len(x) for x in dbf.phases[phase].constituents])
        current_phase_indices = (data.Phase.values == phase)
        if ~np.any(current_phase_indices):
            continue
        points = data.Y.values[np.nonzero(current_phase_indices)][..., :dof]
        statevar_indices = np.nonzero(current_phase_indices)[:len(indep_vals)]
        statevars = {key: np.take(np.asarray(vals), idx)
                     for key, vals, idx in zip(indep_vars, indep_vals, statevar_indices)}
        statevars.update(kwargs)
        if statevars.get('mode', None) is None:
            statevars['mode'] = 'numpy'
        calcres = calculate(dbf, comps, [phase], output=output, points=points, broadcast=False, **statevars)
        result[output].values[np.nonzero(current_phase_indices)] = calcres[output].values
    if not per_phase:
        result[output] = (result[output] * data['NP']).sum(dim='vertex', skipna=True)
    else:
        result['Phase'] = data['Phase'].copy()
        result['NP'] = data['NP'].copy()
    return result


def equilibrium(dbf, comps, phases, conditions, output=None, model=None,
                verbose=False, pbar=True, broadcast=True, calc_opts=None, return_grids=False, **kwargs):
    """
    Calculate the equilibrium state of a system containing the specified
    components and phases, under the specified conditions.

    Parameters
    ----------
    dbf : Database
        Thermodynamic database containing the relevant parameters.
    comps : list
        Names of components to consider in the calculation.
    phases : list or dict
        Names of phases to consider in the calculation.
    conditions : dict or (list of dict)
        StateVariables and their corresponding value.
    output : str or list of str, optional
        Additional equilibrium model properties (e.g., CPM, HM, etc.) to compute.
        These must be defined as attributes in the Model class of each phase.
    model : Model, a dict of phase names to Model, or a seq of both, optional
        Model class to use for each phase.
    verbose : bool, optional
        Print details of calculations. Useful for debugging.
    pbar : bool, optional
        Show a progress bar.
    broadcast : bool
        If True, broadcast conditions against each other. This will compute all combinations.
        If False, each condition should be an equal-length list (or single-valued).
        Disabling broadcasting is useful for calculating equilibrium at selected conditions,
        when those conditions don't comprise a grid.
    calc_opts : dict, optional
        Keyword arguments to pass to `calculate`, the energy/property calculation routine.
    return_grids : bool, optional
        If True, return a tuple of (equilibrium result, grids), where grids is a list of
        Datasets containing the global point set at each hull iteration.
        Mainly useful for teaching and debugging.

    Returns
    -------
    Structured equilibrium calculation.

    Examples
    --------
    None yet.
    """
    if not broadcast:
        raise NotImplementedError('Broadcasting cannot yet be disabled')
    from pycalphad import __version__ as pycalphad_version
    active_phases = unpack_phases(phases) or sorted(dbf.phases.keys())
    comps = sorted(comps)
    if len(set(comps) - set(dbf.elements)) > 0:
        raise EquilibriumError('Components not found in database: {}'.format(','.join(set(comps) - set(dbf.elements))))
    indep_vars = ['T', 'P']
    calc_opts = calc_opts if calc_opts is not None else dict()
    model = model if model is not None else Model
    phase_records = dict()
    callable_dict = kwargs.pop('callables', dict())
    grad_callable_dict = kwargs.pop('grad_callables', dict())
    hess_callable_dict = kwargs.pop('hess_callables', dict())
    points_dict = dict()
    maximum_internal_dof = 0
    # Modify conditions values to be within numerical limits, e.g., X(AL)=0
    # Also wrap single-valued conditions with lists
    conds = _adjust_conditions(conditions)
    for cond in conds.keys():
        if isinstance(cond, (v.Composition, v.ChemicalPotential)) and cond.species not in comps:
            raise ConditionError('{} refers to non-existent component'.format(cond))
    str_conds = OrderedDict((str(key), value) for key, value in conds.items())
    indep_vals = list([float(x) for x in np.atleast_1d(val)]
                      for key, val in str_conds.items() if key in indep_vars)
    components = [x for x in sorted(comps) if not x.startswith('VA')]
    # Construct models for each phase; prioritize user models
    models = unpack_kwarg(model, default_arg=Model)
    # for debugging
    if return_grids:
        intermediate_grids = []
    if verbose:
        print('Components:', ' '.join(comps))
        print('Phases:', end=' ')
    for name in progressbar(active_phases, desc='Initialize (1/3)', unit='phase', disable=not pbar):
        mod = models[name]
        if isinstance(mod, type):
            models[name] = mod = mod(dbf, comps, name)
        variables = sorted(mod.energy.atoms(v.StateVariable).union({key for key in conditions.keys() if key in [v.T, v.P]}), key=str)
        site_fracs = sorted(mod.energy.atoms(v.SiteFraction), key=str)
        maximum_internal_dof = max(maximum_internal_dof, len(site_fracs))
        # Extra factor '1e-100...' is to work around an annoying broadcasting bug for zero gradient entries
        #models[name].models['_broadcaster'] = 1e-100 * Mul(*variables) ** 3
        out = models[name].energy
        undefs = list(out.atoms(Symbol) - out.atoms(v.StateVariable))
        for undef in undefs:
            out = out.xreplace({undef: float(0)})
        if (not callable_dict.get(name, False)) or not (grad_callable_dict.get(name, False)) \
            or (not hess_callable_dict.get(name, False)):
            cf, gf, hf = build_functions(out, [v.P, v.T] + site_fracs)
            if callable_dict.get(name, None) is None:
                callable_dict[name] = cf
            if grad_callable_dict.get(name, None) is None:
                grad_callable_dict[name] = gf
            if hess_callable_dict.get(name, None) is None:
                hess_callable_dict[name] = hf

        # Adjust gradient by the approximate chemical potentials
        hyperplane = Add(*[v.MU(i)*mole_fraction(dbf.phases[name], comps, i)
                           for i in comps if i != 'VA'])
        plane_obj, plane_grad, plane_hess = build_functions(hyperplane, [v.MU(i) for i in comps if i != 'VA']+site_fracs)
        mass_obj, mass_grad, mass_hess = build_functions(Add(*[mole_fraction(dbf.phases[name], comps, i)
                                                               for i in comps if i != 'VA']), site_fracs)
        phase_records[name.upper()] = PhaseRecord(variables=variables,
                                                  grad=grad_callable_dict[name],
                                                  hess=hess_callable_dict[name],
                                                  plane_grad=plane_grad,
                                                  plane_hess=plane_hess,
                                                  mass_obj=mass_obj,
                                                  mass_grad=mass_grad,
                                                  mass_hess=mass_hess)
        if verbose:
            print(name, end=' ')
    if verbose:
        print('[done]', end='\n')

    # 'calculate' accepts conditions through its keyword arguments
    grid_opts = calc_opts.copy()
    grid_opts.update({key: value for key, value in str_conds.items() if key in indep_vars})
    if 'pdens' not in grid_opts:
        grid_opts['pdens'] = 100

    coord_dict = str_conds.copy()
    coord_dict['vertex'] = np.arange(len(components))
    grid_shape = np.meshgrid(*coord_dict.values(),
                             indexing='ij', sparse=False)[0].shape
    coord_dict['component'] = components
    if verbose:
        print('Computing initial grid', end=' ')
    # TODO: vectorize this entire calculation over the conditions
    # TODO: Every condition-set should have its own grid
    grid = calculate(dbf, comps, active_phases, output='GM',
                     model=models, callables=callable_dict, fake_points=True, **grid_opts)
    if return_grids:
        intermediate_grids.append(grid)

    if verbose:
        print('[{0} points, {1}]'.format(len(grid.points), sizeof_fmt(grid.nbytes)), end='\n')

    properties = Dataset({'NP': (list(str_conds.keys()) + ['vertex'],
                                 np.empty(grid_shape)),
                          'GM': (list(str_conds.keys()),
                                 np.empty(grid_shape[:-1])),
                          'MU': (list(str_conds.keys()) + ['component'],
                                 np.empty(grid_shape)),
                          'points': (list(str_conds.keys()) + ['vertex'],
                                     np.empty(grid_shape, dtype=np.int))
                          },
                          coords=coord_dict,
                          attrs={'hull_iterations': 1, 'solve_iterations': 0,
                                 'engine': 'pycalphad %s' % pycalphad_version},
                         )
    # Store the potentials from the previous iteration
    current_potentials = properties.MU.copy()

    convex_progress = progressbar(range(MAX_SEARCH_ITERATIONS), desc='Global Search (2/3)', disable=not pbar)
    for iteration in convex_progress:
        if verbose:
            print('Computing convex hull [iteration {}]'.format(properties.attrs['hull_iterations']))
        # lower_convex_hull will modify properties
        lower_convex_hull(grid, properties, verbose=verbose)
        progress = np.abs(current_potentials - properties.MU).values
        converged = (progress < MIN_SEARCH_PROGRESS).all(axis=-1)
        if verbose:
            print('progress', progress.max(), '[{} conditions updated]'.format(np.sum(~converged)))
        if progress.max() < MIN_SEARCH_PROGRESS:
            if verbose:
                print('Global search complete')
            convex_progress.close()
            break
        current_potentials[...] = properties.MU.values
        if verbose:
            print('Refining convex hull')
        # Insert extra dimensions for non-T,P conditions so GM broadcasts correctly
        energy_broadcast_shape = grid.GM.values.shape[:len(indep_vals)] + \
            (1,) * (len(str_conds) - len(indep_vals)) + (grid.GM.values.shape[-1],)
        driving_forces = np.einsum('...i,...i',
                                   properties.MU.values[..., np.newaxis, :].astype(np.float),
                                   grid.X.values[np.index_exp[...] +
                                                 (np.newaxis,) * (len(str_conds) - len(indep_vals)) +
                                                 np.index_exp[:, :]].astype(np.float)) - \
            grid.GM.values.view().reshape(energy_broadcast_shape)

        for name in active_phases:
            dof = len(models[name].energy.atoms(v.SiteFraction))
            current_phase_indices = (grid.Phase.values == name).reshape(energy_broadcast_shape[:-1] + (-1,))
            # Broadcast to capture all conditions
            current_phase_indices = np.broadcast_arrays(current_phase_indices,
                                                        np.empty(driving_forces.shape))[0]
            # This reshape is safe as long as phases have the same number of points at all indep. conditions
            current_phase_driving_forces = driving_forces[current_phase_indices].reshape(
                current_phase_indices.shape[:-1] + (-1,))
            # Note: This works as long as all points are in the same phase order for all T, P
            current_site_fractions = grid.Y.values[..., current_phase_indices[(0,) * len(str_conds)], :]
            if np.sum(current_site_fractions[(0,) * len(indep_vals)][..., :dof]) == dof:
                # All site fractions are 1, aka zero internal degrees of freedom
                # Impossible to refine these points, so skip this phase
                points_dict[name] = current_site_fractions[(0,) * len(indep_vals)][..., :dof]
                continue
            # Find the N points with largest driving force for a given set of conditions
            # Remember that driving force has a sign, so we want the "most positive" values
            # N is the number of components, in this context
            # N points define a 'best simplex' for every set of conditions
            # We also need to restrict ourselves to one phase at a time
            num_kept_points = 2*len(components)
            trial_indices = np.argpartition(current_phase_driving_forces,
                                            -num_kept_points, axis=-1)[..., -num_kept_points:]
            trial_indices = trial_indices.ravel()
            statevar_indices = np.unravel_index(np.arange(np.multiply.reduce(properties.GM.values.shape + (num_kept_points,))),
                                                properties.GM.values.shape + (num_kept_points,))[:len(indep_vals)]
            points = current_site_fractions[np.index_exp[statevar_indices + (trial_indices,)]]
            points.shape = properties.points.shape[:-1] + (-1, maximum_internal_dof)
            # The Y arrays have been padded, so we should slice off the padding
            points = points[..., :dof]
            starting_points = points.copy()
            #print('Starting points shape: ', points.shape)
            #print(points)
            if len(points) == 0:
                if name in points_dict:
                    del points_dict[name]
                # No nearly stable points: skip this phase
                continue

            num_vars = len(phase_records[name].variables)
            plane_grad = phase_records[name].plane_grad
            plane_hess = phase_records[name].plane_hess
            statevar_grid = np.meshgrid(*itertools.chain(indep_vals), sparse=True, indexing='ij')
            # TODO: A more sophisticated treatment of constraints
            num_constraints = len(dbf.phases[name].sublattices)
            constraint_jac = np.zeros((num_constraints, num_vars-len(indep_vars)))
            # Independent variables are always fixed (in this limited implementation)
            #for idx in range(len(indep_vals)):
            #    constraint_jac[idx, idx] = 1
            # This is for site fraction balance constraints
            var_idx = 0#len(indep_vals)
            for idx in range(len(dbf.phases[name].sublattices)):
                active_in_subl = set(dbf.phases[name].constituents[idx]).intersection(comps)
                constraint_jac[idx,
                               var_idx:var_idx + len(active_in_subl)] = 1
                var_idx += len(active_in_subl)

            newton_iteration = 0
            new_grad = None
            new_cast_grad = None
            new_flattened_points = None
            while newton_iteration < MAX_NEWTON_ITERATIONS:
                if new_grad is None:
                    flattened_points = points.reshape(points.shape[:len(indep_vals)] + (-1, points.shape[-1]))
                    grad_args = itertools.chain([i[..., None] for i in statevar_grid],
                                                [flattened_points[..., i] for i in range(flattened_points.shape[-1])])
                    grad = np.array(phase_records[name].grad(*grad_args), dtype=np.float)
                    # Remove derivatives wrt T,P
                    grad = grad[..., len(indep_vars):]
                    grad.shape = points.shape
                else:
                    grad = new_grad
                    flattened_points = new_flattened_points
                grad[np.isnan(grad).any(axis=-1)] = 0  # This is necessary for gradients on the edge of space
                hess_args = itertools.chain([i[..., None] for i in statevar_grid],
                                            [flattened_points[..., i] for i in range(flattened_points.shape[-1])])
                hess = np.array(phase_records[name].hess(*hess_args), dtype=np.float)
                # Remove derivatives wrt T,P
                hess = hess[..., len(indep_vars):, len(indep_vars):]
                hess.shape = points.shape + (hess.shape[-1],)
                hess[np.isnan(hess).any(axis=(-2, -1))] = np.eye(hess.shape[-1])
                plane_args = itertools.chain([properties.MU.values[..., i][..., None] for i in range(properties.MU.shape[-1])],
                                             [points[..., i] for i in range(points.shape[-1])])
                if new_cast_grad is None:
                    cast_grad = np.array(plane_grad(*plane_args), dtype=np.float)
                    # Remove derivatives wrt chemical potentials
                    cast_grad = cast_grad[..., properties.MU.shape[-1]:]
                else:
                    cast_grad = new_cast_grad
                grad = grad - cast_grad
                plane_args = itertools.chain([properties.MU.values[..., i][..., None] for i in range(properties.MU.shape[-1])],
                                             [points[..., i] for i in range(points.shape[-1])])
                cast_hess = np.array(plane_hess(*plane_args), dtype=np.float)
                # Remove derivatives wrt chemical potentials
                cast_hess = cast_hess[..., properties.MU.shape[-1]:, properties.MU.shape[-1]:]
                cast_hess = -cast_hess + hess
                hess = cast_hess.astype(np.float, copy=False)
                try:
                    e_matrix = np.linalg.inv(hess)
                except np.linalg.LinAlgError:
                    print(hess)
                    print(points)
                    raise
                current = calculate(dbf, comps, name, output='GM',
                                    model=models, callables=callable_dict,
                                    fake_points=False,
                                    points=points.reshape(points.shape[:len(indep_vals)] + (-1, points.shape[-1])),
                                    **grid_opts)
                current_plane = np.multiply(current.X.values.reshape(points.shape[:-1] + (len(components),)),
                                            properties.MU.values[..., np.newaxis, :]).sum(axis=-1)
                current_df = current.GM.values.reshape(points.shape[:-1]) - current_plane
                #print('Inv hess check: ', np.isnan(e_matrix).any())
                #print('grad check: ', np.isnan(grad).any())
                dy_unconstrained = -np.einsum('...ij,...j->...i', e_matrix, grad)
                #print('dy_unconstrained check: ', np.isnan(dy_unconstrained).any())
                proj_matrix = np.dot(e_matrix, constraint_jac.T)
                inv_matrix = np.rollaxis(np.dot(constraint_jac, proj_matrix), 0, -1)
                inv_term = np.linalg.inv(inv_matrix)
                #print('inv_term check: ', np.isnan(inv_term).any())
                first_term = np.einsum('...ij,...jk->...ik', proj_matrix, inv_term)
                #print('first_term check: ', np.isnan(first_term).any())
                # Normally a term for the residual here
                # We only choose starting points which obey the constraints, so r = 0
                cons_summation = np.einsum('...i,...ji->...j', dy_unconstrained, constraint_jac)
                #print('cons_summation check: ', np.isnan(cons_summation).any())
                cons_correction = np.einsum('...ij,...j->...i', first_term, cons_summation)
                #print('cons_correction check: ', np.isnan(cons_correction).any())
                dy_constrained = dy_unconstrained - cons_correction
                #print('dy_constrained check: ', np.isnan(dy_constrained).any())
                # TODO: Support for adaptive changing independent variable steps
                new_direction = dy_constrained
                #print('new_direction', new_direction)
                #print('points', points)
                # Backtracking line search
                if np.isnan(new_direction).any():
                    print('new_direction', new_direction)
                #print('Convergence angle:', -(grad*new_direction).sum(axis=-1) / (np.linalg.norm(grad, axis=-1) * np.linalg.norm(new_direction, axis=-1)))
                new_points = points + INITIAL_STEP_SIZE * new_direction
                alpha = np.full(new_points.shape[:-1], INITIAL_STEP_SIZE, dtype=np.float)
                #alpha[np.all(np.linalg.norm(new_direction, axis=-1) < MIN_DIRECTION_NORM, axis=-1)] = 0
                negative_points = np.any(new_points < 0., axis=-1)
                while np.any(negative_points):
                    alpha[negative_points] *= 0.5
                    new_points = points + alpha[..., np.newaxis] * new_direction
                    negative_points = np.any(new_points < 0., axis=-1)
                # Backtracking line search
                # alpha now contains maximum possible values that keep us inside the space
                # but we don't just want to take the biggest step; we want the biggest step which reduces energy
                new_points = new_points.reshape(new_points.shape[:len(indep_vals)] + (-1, new_points.shape[-1]))
                candidates = calculate(dbf, comps, name, output='GM',
                                       model=models, callables=callable_dict,
                                       fake_points=False, points=new_points, **grid_opts)
                candidate_plane = np.multiply(candidates.X.values.reshape(points.shape[:-1] + (len(components),)),
                                              properties.MU.values[..., np.newaxis, :]).sum(axis=-1)
                energy_diff = (candidates.GM.values.reshape(new_direction.shape[:-1]) - candidate_plane) - current_df
                new_flattened_points = new_points.reshape(new_points.shape[:len(indep_vals)] + (-1, new_points.shape[-1]))
                new_grad_args = itertools.chain([i[..., None] for i in statevar_grid],
                                            [new_flattened_points[..., i] for i in range(new_flattened_points.shape[-1])])
                new_grad = np.array(phase_records[name].grad(*new_grad_args), dtype=np.float)
                # Remove derivatives wrt T,P
                new_grad = new_grad[..., len(indep_vars):]
                new_points.shape = new_direction.shape
                new_grad.shape = new_points.shape
                new_plane_args = itertools.chain([properties.MU.values[..., i][..., None] for i in range(properties.MU.shape[-1])],
                                             [new_points[..., i] for i in range(new_points.shape[-1])])
                new_cast_grad = np.array(plane_grad(*new_plane_args), dtype=np.float)
                # Remove derivatives wrt chemical potentials
                new_cast_grad = new_cast_grad[..., properties.MU.shape[-1]:]
                new_grad = new_grad - new_cast_grad
                bad_steps = energy_diff > alpha * 1e-4 * (new_direction * grad).sum(axis=-1)
                bad_steps |= np.abs(np.multiply(new_direction, new_grad).sum(axis=-1)) > 0.9*np.abs(np.multiply(new_direction, grad).sum(axis=-1))
                backtracking_iterations = 0
                while np.any(bad_steps):
                    alpha[bad_steps] *= 0.9
                    new_points = points + alpha[..., np.newaxis] * new_direction
                    #print('new_points', new_points)
                    #print('bad_steps', bad_steps)
                    new_points = new_points.reshape(new_points.shape[:len(indep_vals)] + (-1, new_points.shape[-1]))
                    candidates = calculate(dbf, comps, name, output='GM',
                                           model=models, callables=callable_dict,
                                           fake_points=False, points=new_points, **grid_opts)
                    candidate_plane = np.multiply(candidates.X.values.reshape(points.shape[:-1] + (len(components),)),
                                                  properties.MU.values[..., np.newaxis, :]).sum(axis=-1)
                    energy_diff = (candidates.GM.values.reshape(new_direction.shape[:-1]) - candidate_plane) - current_df
                    new_flattened_points = new_points.reshape(new_points.shape[:len(indep_vals)] + (-1, new_points.shape[-1]))
                    new_grad_args = itertools.chain([i[..., None] for i in statevar_grid],
                                                [new_flattened_points[..., i] for i in range(new_flattened_points.shape[-1])])
                    new_grad = np.array(phase_records[name].grad(*new_grad_args), dtype=np.float)
                    # Remove derivatives wrt T,P
                    new_grad = new_grad[..., len(indep_vars):]
                    new_points.shape = new_direction.shape
                    new_grad.shape = new_points.shape
                    new_plane_args = itertools.chain([properties.MU.values[..., i][..., None] for i in range(properties.MU.shape[-1])],
                                                 [new_points[..., i] for i in range(new_points.shape[-1])])
                    new_cast_grad = np.array(plane_grad(*new_plane_args), dtype=np.float)
                    # Remove derivatives wrt chemical potentials
                    new_cast_grad = new_cast_grad[..., properties.MU.shape[-1]:]
                    new_grad = new_grad - new_cast_grad
                    bad_steps = energy_diff > alpha * 1e-4 * (new_direction * grad).sum(axis=-1)
                    bad_directions = (new_direction * grad).sum(axis=-1) > 0
                    #if np.any(bad_directions):
                    #    print('bad points', new_points[np.nonzero(bad_directions)])
                    #    print('bad directions', new_direction[np.nonzero(bad_directions)])
                    #    print('bad gradients', grad[np.nonzero(bad_directions)])
                    #    print('energy_diff', energy_diff[np.nonzero(bad_directions)])
                    bad_steps |= np.abs(np.multiply(new_direction, new_grad).sum(axis=-1)) > 0.9*np.abs(np.multiply(new_direction, grad).sum(axis=-1))
                    backtracking_iterations += 1
                    if backtracking_iterations > MAX_BACKTRACKING:
                        # Couldn't find a suitable step -- take no step for bad steps
                        print('Backtracking failed')
                        new_grad[np.nonzero(~bad_steps)] = grad[np.nonzero(~bad_steps)]
                        new_points[np.nonzero(~bad_steps)] = points[np.nonzero(~bad_steps)]
                        break
                    alpha[alpha < MIN_SOLVE_ALPHA] = 0
                #print('backtracking iterations', backtracking_iterations)
                biggest_step = np.max(np.linalg.norm(new_points - points, axis=-1))
                if biggest_step < MIN_DIRECTION_NORM:
                    if verbose:
                        print('N-R convergence on mini-iteration', newton_iteration, '[{}]'.format(name))
                    points = new_points
                    break
                if verbose:
                    #print('Biggest step:', biggest_step)
                    #print('points', points)
                    #print('grad of points', grad)
                    #print('new_direction', new_direction)
                    #print('alpha', alpha)
                    #print('new_points', new_points)
                    pass
                points = new_points
                newton_iteration += 1
            new_points = points.reshape(points.shape[:len(indep_vals)] + (-1, points.shape[-1]))
            print('points.shape', points.shape)
            print('starting_points.shape', starting_points.shape)
            starting_points.shape = points.shape[:len(indep_vals)] + (-1, points.shape[-1])
            new_points = np.concatenate((starting_points, new_points), axis=-2)
            points_dict[name] = new_points

        if verbose:
            print('Rebuilding grid', end=' ')
        grid = calculate(dbf, comps, active_phases, output='GM',
                         model=models, callables=callable_dict,
                         fake_points=True, points=points_dict, **grid_opts)
        if return_grids:
            intermediate_grids.append(grid)
        if verbose:
            print('[{0} points, {1}]'.format(len(grid.points), sizeof_fmt(grid.nbytes)), end='\n')
        properties.attrs['hull_iterations'] += 1
    # Make sure all the verbose output appears to the user
    if verbose:
        print('Refining equilibrium')
        sys.stdout.flush()
    # One last call to ensure 'properties' and 'grid' are consistent with one another
    lower_convex_hull(grid, properties, verbose=verbose)
    indexer = []
    for idx, vals in enumerate(indep_vals):
        indexer.append(np.arange(len(vals), dtype=np.int)[idx * (np.newaxis,) + np.index_exp[:] + \
                                                          (len(conds.keys())-idx+1) * (np.newaxis,)])
    indexer.append(properties['points'].values[..., np.newaxis])
    indexer.append(np.arange(grid['X'].values.shape[-1], dtype=np.int)[(len(conds.keys())) * (np.newaxis,) + np.index_exp[:]])
    ravelled_X_view = grid['X'].values[tuple(indexer)]
    indexer[-1] = np.arange(grid['Y'].values.shape[-1], dtype=np.int)[(len(conds.keys())) * (np.newaxis,) + np.index_exp[:]]
    ravelled_Y_view = grid['Y'].values[tuple(indexer)]
    indexer = []
    for idx, vals in enumerate(indep_vals):
        indexer.append(np.arange(len(vals), dtype=np.int)[idx * (np.newaxis,) + np.index_exp[:] + \
                                                          (len(conds.keys())-idx) * (np.newaxis,)])
    indexer.append(properties['points'].values)
    ravelled_Phase_view = grid['Phase'].values[tuple(indexer)]
    # Copy final point values from the grid and drop the index array
    # For some reason direct construction doesn't work. We have to create empty and then assign.
    properties['X'] = DataArray(np.empty_like(ravelled_X_view),
                                dims=properties['points'].dims + ('component',))
    properties['X'].values[...] = ravelled_X_view
    properties['Y'] = DataArray(np.empty_like(ravelled_Y_view),
                                dims=properties['points'].dims + ('internal_dof',))
    properties['Y'].values[...] = ravelled_Y_view
    # TODO: What about invariant reactions? We should perform a final driving force calculation here.
    # We can handle that in the same post-processing step where we identify single-phase regions.
    properties['Phase'] = DataArray(np.empty_like(ravelled_Phase_view),
                                    dims=properties['points'].dims)
    properties['Phase'].values[...] = ravelled_Phase_view
    del properties['points']
    it = np.nditer(properties['GM'].values, flags=['multi_index'])
    num_conds = np.prod([len(x) for x in properties['GM'].coords.values()])
    # For single-condition calculations, don't generate a separate progress bar
    multi_phase_progress = progressbar(desc='Solve (3/3)', total=num_conds,
                                       unit='cond', disable=(num_conds==1) or (not pbar))
    while not it.finished:
        # A lot of this code relies on cur_conds being ordered!
        cur_conds = OrderedDict(zip(properties['GM'].coords.keys(),
                                    [b[a] for a, b in zip(it.multi_index, properties['GM'].coords.values())]))
        # sum of independently specified components
        indep_sum = np.sum([float(val) for i, val in cur_conds.items() if i.startswith('X_')])
        if indep_sum > 1:
            # Sum of independent component mole fractions greater than one
            # Skip this condition set
            # We silently allow this to make 2-D composition mapping easier
            properties['MU'].values[it.multi_index] = np.nan
            properties['NP'].values[it.multi_index + np.index_exp[:len(phases)]] = np.nan
            properties['X'].values[it.multi_index + np.index_exp[:len(phases)]] = np.nan
            properties['Y'].values[it.multi_index] = np.nan
            properties['GM'].values[it.multi_index] = np.nan
            it.iternext()
            continue
        dependent_comp = set(comps) - set([i[2:] for i in cur_conds.keys() if i.startswith('X_')]) - {'VA'}
        if len(dependent_comp) == 1:
            dependent_comp = list(dependent_comp)[0]
        else:
            multi_phase_progress.close()
            raise ValueError('Number of dependent components different from one')
        #chem_pots = OrderedDict(zip(properties.coords['component'].values, properties['MU'].values[it.multi_index]))
        # For single-condition calculations, make a progress bar for the individual solve
        solve_progress = progressbar(range(MAX_SOLVE_ITERATIONS), desc='Solve (3/3',
                                     disable=(num_conds>1) or (not pbar))
        for cur_iter in solve_progress:
            #print('CUR_ITER:', cur_iter)
            phases = list(properties['Phase'].values[it.multi_index])
            if '' in phases:
                old_phase_length = phases.index('')
            else:
                old_phase_length = -1
            remove_degenerate_phases(properties, it.multi_index)
            # TODO: This is a temporary workaround to allow phase diagram calculation to work decently
            if kwargs.get('_approx', None) is not None:
                break
            phases = list(properties['Phase'].values[it.multi_index])
            if '' in phases:
                new_phase_length = phases.index('')
            else:
                new_phase_length = -1
            # Are there removed phases?
            if '' in phases:
                num_phases = phases.index('')
            else:
                num_phases = len(phases)
            zero_dof = np.all((properties['Y'].values[it.multi_index] == 1.) | np.isnan(properties['Y'].values[it.multi_index]))
            if (num_phases == 1) and zero_dof:
                # Single phase with zero internal degrees of freedom, can't do any refinement
                # TODO: In the future we may be able to refine other degrees of freedom like temperature
                # Chemical potentials have no meaning for this case
                properties['MU'].values[it.multi_index] = np.nan
                break
            phases = properties['Phase'].values[it.multi_index + np.index_exp[:num_phases]]
            #num_sitefrac_bals = sum([len(dbf.phases[i].sublattices) for i in phases])
            #num_mass_bals = len([i for i in cur_conds.keys() if i.startswith('X_')]) + 1
            phase_fracs = properties['NP'].values[it.multi_index + np.index_exp[:len(phases)]]
            phase_dof = [len(set(phase_records[name].variables)-{v.T, v.P}) for name in phases]
            # Flatten site fractions array and remove nan padding
            site_fracs = properties['Y'].values[it.multi_index].ravel()
            # That *should* give us the internal dof
            # This may break if non-padding nan's slipped in from elsewhere...
            site_fracs = site_fracs[~np.isnan(site_fracs)]
            site_fracs[site_fracs < MIN_SITE_FRACTION] = MIN_SITE_FRACTION
            phase_fracs[phase_fracs < MIN_SITE_FRACTION] = MIN_SITE_FRACTION
            var_idx = 0
            for name in phases:
                for idx in range(len(dbf.phases[name].sublattices)):
                    active_in_subl = set(dbf.phases[name].constituents[idx]).intersection(comps)
                    site_fracs[var_idx:var_idx + len(active_in_subl)] /= \
                        np.sum(site_fracs[var_idx:var_idx +len(active_in_subl)], keepdims=True)
                    var_idx += len(active_in_subl)

            l_constraints, constraint_jac, l_multipliers, old_chem_pots, mole_fraction_funcs = \
                _compute_constraints(dbf, comps, phases, cur_conds, site_fracs, phase_fracs, phase_records,
                                     chem_pots=properties['MU'].values[it.multi_index].copy())
            num_vars = len(site_fracs) + len(phases)
            l_hessian, gradient_term = _build_multiphase_system(dbf, comps, phases, cur_conds, site_fracs, phase_fracs,
                                                                l_constraints, constraint_jac, l_multipliers,
                                                                callable_dict, phase_records)
            try:
                step = np.linalg.solve(l_hessian, gradient_term)
            except np.linalg.LinAlgError:
                print('Failed to compute ', cur_conds)
                properties['GM'].values[it.multi_index] = np.nan
                solve_progress.close()
                break
            if np.any(np.isnan(step)):
                print('PHASES: ', phases)
                print('SITE FRACTIONS: ', site_fracs)
                print('PHASE FRACTIONS: ', phase_fracs)
                print('HESSIAN: ', l_hessian)
                solve_progress.close()
                print('Bad step: '+str(step))
                break
            # Backtracking line search
            # First restrict alpha to steps in the feasible region
            alpha = 1
            #while ((np.any((site_fracs + alpha * step[:len(site_fracs)]) < 0.1*MIN_SITE_FRACTION) or
            #       np.any((phase_fracs + alpha * step[len(site_fracs):len(site_fracs)+len(phases)]) < 0)) and
            #       alpha > MIN_SOLVE_ALPHA):
            #    alpha *= 0.999
            #if alpha <= MIN_SOLVE_ALPHA:
            #    alpha = 0
            #print('INITIAL ALPHA', alpha)
            #print('STEP', step)
            #print('SITE FRACS', site_fracs)
            #print('PHASE FRACS', phase_fracs)
            # Take the largest step which reduces the energy
            old_energy = copy.deepcopy(properties['GM'].values[it.multi_index])
            #print('OLD ENERGY', old_energy)
            # XXX: Why is it necessary to square this?
            old_constrained_objective = old_energy + ((l_multipliers * l_constraints).sum()) ** 2
            #print('OLD OBJ', old_constrained_objective)
            old_chem_pots = properties['MU'].values[it.multi_index].copy()
            #print('STARTING ALPHA', alpha)
            wolfe_conditions = False
            while alpha > MIN_SOLVE_ALPHA:
                #print('ALPHA', alpha)
                candidate_site_fracs = site_fracs + alpha * step[:len(site_fracs)]
                candidate_site_fracs[candidate_site_fracs < MIN_SITE_FRACTION] = MIN_SITE_FRACTION
                candidate_site_fracs[candidate_site_fracs > 1] = 1
                candidate_l_multipliers = l_multipliers + alpha * step[num_vars:]
                #print('CANDIDATE L MULTIPLIERS', candidate_l_multipliers)
                candidate_phase_fracs = phase_fracs + \
                                       alpha * step[len(candidate_site_fracs):len(candidate_site_fracs)+len(phases)]
                candidate_phase_fracs[candidate_phase_fracs < MIN_SITE_FRACTION] = MIN_SITE_FRACTION
                candidate_phase_fracs[candidate_phase_fracs > 1] = 1
                #if len(phases) == len([c for c in comps if c != 'VA']):
                #    # We have the maximum number of phases
                #    # Compute the chemical potentials exactly from the tangent hyperplane
                #    phase_compositions = np.zeros((len(phases), len([c for c in comps if c != 'VA'])))
                #    phase_energies = np.zeros((len(phases)))
                #    var_offset = 0
                #    for phase_idx in range(len(phases)):
                #        for comp_idx, comp in enumerate([c for c in comps if c != 'VA']):
                #            phase_compositions[phase_idx, comp_idx] = \
                #                mole_fraction_funcs[(phases[phase_idx], comp)][0](*candidate_site_fracs[var_offset:var_offset+phase_dof[phase_idx]])
                #        phase_energies[phase_idx] = callable_dict[phases[phase_idx]](
                #            *itertools.chain([cur_conds['P'], cur_conds['T']],
                #                             candidate_site_fracs[var_offset:var_offset + phase_dof[phase_idx]]))
                #        var_offset += phase_dof[phase_idx]
                #    exact_chempots = np.linalg.solve(phase_compositions, phase_energies)
                #else:
                #    exact_chempots = None
                (candidate_l_constraints, candidate_constraint_jac, candidate_l_multipliers, candidate_chem_pots,
                mole_fraction_funcs) = \
                    _compute_constraints(dbf, comps, phases, cur_conds,
                                         candidate_site_fracs, candidate_phase_fracs, phase_records,
                                         l_multipliers=candidate_l_multipliers)
                #print('CANDIDATE L MULTIPLIERS AFTER', candidate_l_multipliers)
                #print('CANDIDATE_L_CONSTRAINTS', candidate_l_constraints)
                #print('CANDIDATE L MULS', candidate_l_constraints * candidate_l_multipliers)
                #print('CANDIDATE L MUL SUM', (candidate_l_multipliers *
                #                                    candidate_l_constraints).sum()
                #      )
                candidate_energy = _compute_multiphase_objective(dbf, comps, phases, cur_conds, candidate_site_fracs,
                                                                 candidate_phase_fracs,
                                                                 callable_dict)
                # XXX: Why is it necessary to square this?
                candidate_constrained_objective = candidate_energy + ((candidate_l_multipliers *
                                                                       candidate_l_constraints).sum()) ** 2
                #print('CANDIDATE CHEM POTS', candidate_chem_pots)
                #print('CANDIDATE ENERGY', candidate_energy)
                #print('CANDIDATE OBJ', candidate_constrained_objective)
                #print('STEP', step[:num_vars])
                #print('GRADIENT TERM', gradient_term)
                #print('CANDIDATE GRADIENT', candidate_gradient_term)

                # Remember that 'gradient_term' is actually the NEGATIVE of the gradient
                wolfe_conditions = (candidate_constrained_objective - old_constrained_objective) <= \
                                   alpha * 1e-4 * (step * -gradient_term).sum(axis=-1)
                #print('WOLFE CONDITION 1', wolfe_conditions)
                # Optimization to avoid costly gradient calculation if Wolfe conditions won't be met anyway
                if wolfe_conditions:
                    candidate_gradient_term = _build_multiphase_gradient(dbf, comps, phases,
                                                                         cur_conds, candidate_site_fracs,
                                                                         candidate_phase_fracs,
                                                                         l_constraints, constraint_jac,
                                                                         l_multipliers, callable_dict, phase_records)
                    #print('CANDIDATE GRAD SUM', np.multiply(step[:num_vars], candidate_gradient_term[:num_vars]).sum())
                    #print('OLD GRAD SUM', np.multiply(step[:num_vars], gradient_term[:num_vars]).sum())
                    wolfe_conditions &= np.abs(np.multiply(step, -candidate_gradient_term).sum(axis=-1)) <= \
                                        0.9*np.abs(np.multiply(step, -gradient_term).sum(axis=-1))
                # Seems to be necessary for some unit tests to explicitly allow chemical potential updates
                chempot_update = (candidate_constrained_objective - old_constrained_objective) <= MIN_SOLVE_ENERGY_PROGRESS
                chempot_update &= np.abs(candidate_chem_pots - old_chem_pots).max() > 0.1
                wolfe_conditions |= chempot_update
                #print('WOLFE CONDITION 1&2', wolfe_conditions)
                if wolfe_conditions:
                    break
                alpha *= 0.5
            #print('RESULT ALPHA', alpha)
            #print('WOLFE CONDITIONS', wolfe_conditions)
            if wolfe_conditions:
                # We updated degrees of freedom this iteration
                properties['MU'].values[it.multi_index] = candidate_chem_pots
                properties['NP'].values[it.multi_index + np.index_exp[:len(phases)]] = candidate_phase_fracs
                properties['X'].values[it.multi_index + np.index_exp[:len(phases)]] = 0
                properties['GM'].values[it.multi_index] = candidate_energy
                var_offset = 0
                for phase_idx in range(len(phases)):
                    properties['Y'].values[it.multi_index + np.index_exp[phase_idx, :phase_dof[phase_idx]]] = \
                        candidate_site_fracs[var_offset:var_offset+phase_dof[phase_idx]]
                    for comp_idx, comp in enumerate([c for c in comps if c != 'VA']):
                        properties['X'].values[it.multi_index + np.index_exp[phase_idx, comp_idx]] = \
                            mole_fraction_funcs[(phases[phase_idx], comp)][0](*candidate_site_fracs[var_offset:var_offset+phase_dof[phase_idx]])
                    var_offset += phase_dof[phase_idx]

            properties.attrs['solve_iterations'] += 1
            if verbose:
                print('Chem pot progress', np.linalg.norm(properties['MU'].values[it.multi_index] - old_chem_pots, ord=np.inf))
                print('Energy progress', properties['GM'].values[it.multi_index] - old_energy)
            no_progress = ~wolfe_conditions
            #no_progress = np.abs(properties['MU'].values[it.multi_index] - old_chem_pots).max() < 0.1
            #no_progress &= np.abs(properties['GM'].values[it.multi_index] - old_energy) < MIN_SOLVE_ENERGY_PROGRESS
            if no_progress:
                if verbose:
                    print('No progress')
                solve_progress.close()
                break
        multi_phase_progress.update()
        it.iternext()
    multi_phase_progress.close()

    # Compute equilibrium values of any additional user-specified properties
    output = output if isinstance(output, (list, tuple, set)) else [output]
    # We already computed these properties so don't recompute them
    output = sorted(set(output) - {'GM', 'MU'})
    for out in output:
        if (out is None) or (len(out) == 0):
            continue
        # TODO: How do we know if a specified property should be per_phase or not?
        # For now, we make a best guess
        if (out == 'degree_of_ordering') or (out == 'DOO'):
            per_phase = True
        else:
            per_phase = False
        properties.merge(_eqcalculate(dbf, comps, active_phases, conditions, out,
                                      data=properties, per_phase=per_phase, **calc_opts), inplace=True, compat='equals')
    properties.attrs['created'] = datetime.utcnow()
    if return_grids:
        return properties, intermediate_grids
    else:
        return properties



"""
The plot utils module contains some useful routines related to plotting.
"""

import matplotlib.patches as mpatches
import numpy as np

def phase_legend(phases):
    """
    Build matplotlib handles for the plot legend.

    Parameters
    ----------
    phases : list
        Names of the phases.

    Returns
    -------
    A tuple containing:
    (1) A list of matplotlib handle objects
    (2) A dict mapping phase names to their RGB color on the plot

    Examples
    --------
    >>> legend_handles, colors = phase_legend(['FCC_A1', 'BCC_A2', 'LIQUID'])
    """
    colorlist = {}
    # colors from Junwei Huang, March 21 2013
    # exclude green and red because of their special meaning on the diagram
    colorvalues = ["0000FF", "FFFF00", "FF00FF", "00FFFF", "000000", "800000",
                   "008000", "000080", "808000", "800080", "008080", "808080",
                   "C00000", "00C000", "0000C0", "C0C000", "C000C0", "00C0C0",
                   "C0C0C0", "400000", "004000", "000040", "404000", "400040",
                   "004040", "404040", "200000", "002000", "000020", "202000",
                   "200020", "002020", "202020", "600000", "006000", "000060",
                   "606000", "600060", "006060", "606060", "A00000", "00A000",
                   "0000A0", "A0A000", "A000A0", "00A0A0", "A0A0A0", "E00000",
                   "00E000", "0000E0", "E0E000", "E000E0", "00E0E0", "E0E0E0"]
    mxx = len(colorvalues)
    phasecount = 0
    legend_handles = []
    for phase in phases:
        phase = phase.upper()
        colorlist[phase] = "#"+colorvalues[np.mod(phasecount, mxx)]
        legend_handles.append(mpatches.Patch(color=colorlist[phase],
                                             label=phase))
        phasecount = phasecount + 1
    return legend_handles, colorlist



"""
This module deals with managing the result of an equilibrium calculation.
"""

from operator import itemgetter
from collections import defaultdict
import copy
import pycalphad.variables as v

class SublatticeResult(object):
    def __init__(self):
        self.site_count = None
        self.site_fractions = dict()

    def __repr__(self):
        return '{0!s}({1!r})'.format(self.__class__, self.__dict__)

    def __str__(self):
        return '    '.join(['{0:<} {1:E}'.format(k, v) \
            for k, v in sorted(self.site_fractions.items(), key=itemgetter(0))])

class PhaseResult(object):
    def __init__(self):
        self.name = None
        self.volume_fraction = None
        self.multiplicity = None
        self.sublattices = list()

    @property
    def mole_fractions(self):
        result = defaultdict(lambda: 0.0)
        site_ratio_normalization = 0.0
        # Calculate normalization factor
        for sublattice in self.sublattices:
            if 'VA' in set(sublattice.site_fractions.keys()):
                site_ratio_normalization += sublattice.site_count * \
                    (1.0 - sublattice.site_fractions['VA'])
            else:
                site_ratio_normalization += sublattice.site_count
        site_ratios = [c.site_count/site_ratio_normalization \
            for c in self.sublattices]
        # Sum up site fraction contributions
        for idx, sublattice in enumerate(self.sublattices):
            for component in sublattice.site_fractions.keys():
                if component == 'VA':
                    continue
                result[component] += site_ratios[idx] * \
                    sublattice.site_fractions[component]
        return result

    def __repr__(self):
        return '{0!s}({1!r})'.format(self.__class__, self.__dict__)

    def __str__(self):
        res = '{0:<}#{1}    {2:E}\n'.format(self.name,\
            self.multiplicity, self.volume_fraction)
        res += 'Mole fractions:    '
        res += '    '.join(['{0:<} {1:E}'.format(k, v) \
            for k, v in sorted(self.mole_fractions.items(), key=itemgetter(0))])
        res += '\n'
        for idx, sublattice in enumerate(self.sublattices):
            res += 'Sublattice {0}: {1} sites\n    {2!s}\n'.format(idx+1, \
                sublattice.site_count, sublattice)
        return res

class EquilibriumResult(object):
    def __init__(self, phases, components, potentials, energy, variables):
        self.phases = list()
        self.energy = energy
        self.components = components
        self.potentials = potentials
        phase_res = PhaseResult()
        for variable, value in variables:
            if isinstance(variable, v.PhaseFraction):
                # New phase: Append old one if not empty
                if phase_res.name is not None:
                    self.phases.append(copy.deepcopy(phase_res))
                phase_res = PhaseResult()
                phase_res.name = variable.phase_name
                phase_res.multiplicity = variable.multiplicity
                phase_res.volume_fraction = value
            elif isinstance(variable, v.SiteFraction):
                # Add sublattices if this variable has a larger index
                if variable.sublattice_index >= len(phase_res.sublattices):
                    phase_res.sublattices.extend(SublatticeResult() \
                        for i in range(variable.sublattice_index - \
                            len(phase_res.sublattices) + 1))
                phase_res.sublattices[variable.sublattice_index].site_count = \
                    phases[variable.phase_name].sublattices[
                        variable.sublattice_index]
                phase_res.sublattices[
                    variable.sublattice_index
                    ].site_fractions[variable.species] = value
        # Add final phase
        self.phases.append(copy.deepcopy(phase_res))

    @property
    def mole_fractions(self):
        result = defaultdict(lambda: 0.0)
        # Sum up phase contributions
        for phase in self.phases:
            for component, value in phase.mole_fractions.items():
                result[component] += phase.volume_fraction * value
        return result

    def __repr__(self):
        return '{0!s}({1!r})'.format(self.__class__, self.__dict__)

    def __str__(self):
        res = ''
        res += 'Molar Gibbs Energy:    {0:E}\n'.format(self.energy)
        res += 'Potentials:\n'
        res += '    '.join(['{0!s}={1}'.format(k, v) \
            for k, v in self.potentials.items()])
        res += '\n'
        res += 'Molar Composition:\n'
        res += '    '.join(['X({0!s})={1:E}'.format(k, v) \
            for k, v in sorted(list(self.mole_fractions.items()))])
        res += '\n'
        res += '\n'
        res += '\n'.join([str(ph) for ph in self.phases])
        return res



"""
The binary module enables plotting of binary
isobaric phase diagrams.
"""
import numpy as np

from pycalphad import equilibrium
import pycalphad.variables as v
from pycalphad.plot.eqplot import eqplot


def binplot(dbf, comps, phases, conds, eq_kwargs=None, **plot_kwargs):
    """
    Calculate the binary isobaric phase diagram.
    This function is a convenience wrapper around equilibrium() and eqplot().

    Parameters
    ----------
    dbf : Database
        Thermodynamic database containing the relevant parameters.
    comps : list
        Names of components to consider in the calculation.
    phases : list
        Names of phases to consider in the calculation.
    conds : dict
        Maps StateVariables to values and/or iterables of values.
        For binplot only one changing composition and one potential coordinate each is supported.
    eq_kwargs : optional
        Keyword arguments to equilibrium().
    plot_kwargs : optional
        Keyword arguments to eqplot().

    Returns
    -------
    A phase diagram as a figure.

    Examples
    --------
    None yet.
    """
    eq_kwargs = eq_kwargs if eq_kwargs is not None else dict()
    indep_comp = [key for key, value in conds.items() if isinstance(key, v.Composition) and len(np.atleast_1d(value)) > 1]
    indep_pot = [key for key, value in conds.items() if ((key == v.T) or (key == v.P)) and len(np.atleast_1d(value)) > 1]
    if (len(indep_comp) != 1) or (len(indep_pot) != 1):
        raise ValueError('binplot() requires exactly one composition and one potential coordinate')
    indep_comp = indep_comp[0]
    indep_pot = indep_pot[0]

    full_eq = equilibrium(dbf, comps, phases, conds, **eq_kwargs)
    return eqplot(full_eq, x=indep_comp, y=indep_pot, **plot_kwargs)



import pycalphad.variables as v
from pycalphad.model import Model
from pycalphad.io.database import Database

# Trigger format extension hooks
import pycalphad.io.tdb

from pycalphad.core.calculate import calculate
from pycalphad.core.equilibrium import equilibrium
from pycalphad.core.equilibrium import EquilibriumError, ConditionError
from pycalphad.plot.binary import binplot
from pycalphad.plot.eqplot import eqplot
from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



"""
This module manages interactions with the autograd library.
"""
import autograd.numpy as anp
from autograd import elementwise_grad, jacobian
from sympy import lambdify
from pycalphad.core.utils import NumPyPrinter
from itertools import chain


def elementwise_hess(fun, argnum=0):
    """
    From https://github.com/HIPS/autograd/issues/60
    """
    def sum_latter_dims(x):
        return anp.sum(x.reshape(x.shape[0], -1), 1)

    def sum_grad_output(*args, **kwargs):
        return sum_latter_dims(elementwise_grad(fun)(*args, **kwargs))
    return jacobian(sum_grad_output, argnum)


def build_functions(sympy_graph, variables):
    logical_np = [{'And': anp.logical_and, 'Or': anp.logical_or, 'Abs': anp.abs}, anp]
    obj = lambdify(tuple(variables), sympy_graph, dummify=True,
                   modules=logical_np, printer=NumPyPrinter)

    def argwrapper(args):
        return obj(*args)

    def grad_func(*args):
        inp = anp.array(anp.broadcast_arrays(*args))
        result = anp.atleast_2d(elementwise_grad(argwrapper)(inp))
        # Put 'gradient' axis at end
        axes = list(range(len(result.shape)))
        result = result.transpose(*chain(axes[1:], [axes[0]]))
        return result

    def hess_func(*args):
        result = anp.atleast_3d(elementwise_hess(argwrapper)(anp.array(anp.broadcast_arrays(*args))))
        # Put 'hessian' axes at end
        axes = list(range(len(result.shape)))
        result = result.transpose(*chain(axes[2:], axes[0:2]))
        return result

    return obj, grad_func, hess_func



"""
The constraints module contains definitions for equilibrium constraints and
their Jacobian.
"""

import pycalphad.variables as v
from sympy import S, Add

def mole_fraction(phase, active_comps, species):
    """
    Return a SymPy object representing the mole fraction as a function of
    site fractions.

    Parameters
    ----------
    phase : Phase
        Phase object corresponding to the phase of interest.
    active_comps : list of str
        Names of components to consider.
    species : str
        Names of species to consider.

    Returns
    -------
    SymPy object representing the mole fraction.

    Examples
    --------
    >>> dbf = Database('alfe_sei.TDB')
    >>> mole_fraction(dbf.phases['FCC_A1'], ['AL', 'FE', 'VA'], 'AL')
    """
    result = S.Zero
    site_ratio_normalization = S.Zero
    # Calculate normalization factor
    for idx, sublattice in enumerate(phase.constituents):
        active = set(sublattice).intersection(set(active_comps))
        if 'VA' in active:
            site_ratio_normalization += phase.sublattices[idx] * \
                (1.0 - v.SiteFraction(phase.name, idx, 'VA'))
        else:
            site_ratio_normalization += phase.sublattices[idx]
    site_ratios = [c/site_ratio_normalization for c in phase.sublattices]
    # Sum up site fraction contributions from each sublattice
    for idx, sublattice in enumerate(phase.constituents):
        active = set(sublattice).intersection(set(active_comps))
        if species in active:
            result += site_ratios[idx] * \
                v.SiteFraction(phase.name, idx, species)
    return result




"""
The utils module handles helper routines for equilibrium calculation.
"""
from __future__ import division
import pycalphad.variables as v
from pycalphad.core.halton import halton
from pycalphad.core.constants import MIN_SITE_FRACTION
from sympy.utilities.lambdify import lambdify
from sympy.printing.lambdarepr import LambdaPrinter
import numpy as np
import operator
import functools
import itertools
import collections
from importlib import import_module

_NUMBA = None

try:
    _NUMBA = import_module('numba')

    @_NUMBA.vectorize
    def nbwhere(condition, x, y):
        if condition:
            return x
        else:
            return y
except ImportError:
    pass

try:
    # Only available in numpy 1.10 and newer
    from numpy import broadcast_to
except ImportError:
    def broadcast_to(arr, shape):
        "Broadcast an array to a desired shape. Returns a view."
        return np.broadcast_arrays(arr, np.empty(shape, dtype=np.bool))[0]


class NumPyPrinter(LambdaPrinter):
    """
    Numpy printer which handles vectorized piecewise functions,
    logical operators, etc.
    """
    _default_settings = {
        "order": "none",
        "full_prec": "auto",
    }

    def _print_seq(self, seq, delimiter=', '):
        "General sequence printer: converts to tuple"
        # Print tuples here instead of lists because numba supports
        #     tuples in nopython mode.
        return '({},)'.format(delimiter.join(self._print(item) for item in seq))

    #def _print_Integer(self, expr):
    #    # Upcast to work around an integer overflow bug
    #    return 'asarray({0}, dtype=float)'.format(expr)

    def _print_MatrixBase(self, expr):
        return "%s(%s)" % ('array', self._print(expr.tolist()))

    _print_SparseMatrix = \
        _print_MutableSparseMatrix = \
        _print_ImmutableSparseMatrix = \
        _print_Matrix = \
        _print_DenseMatrix = \
        _print_MutableDenseMatrix = \
        _print_ImmutableMatrix = \
        _print_ImmutableDenseMatrix = \
        _print_MatrixBase

    def _print_MatMul(self, expr):
        "Matrix multiplication printer"
        return '({0})'.format(').dot('.join(self._print(i) for i in expr.args))

    def _print_Piecewise(self, expr):
        "Piecewise function printer"
        exprs = [self._print(arg.expr) for arg in expr.args]
        conds = [self._print(arg.cond) for arg in expr.args]
        if (len(conds) == 1) and exprs[0] == '0':
            return '0'
        # We expect exprs and conds to be in order here
        # First condition to evaluate to True is returned
        # Default result: float zero
        result = 'where({0}, {1}, 0.)'.format(conds[-1], exprs[-1])
        for expr, cond in reversed(list(zip(exprs[:-1], conds[:-1]))):
            result = 'where({0}, {1}, {2})'.format(cond, expr, result)

        return result

    def _print_And(self, expr):
        "Logical And printer"
        # We have to override LambdaPrinter because it uses Python 'and' keyword.
        # If LambdaPrinter didn't define it, we could use StrPrinter's
        # version of the function and add 'logical_and' to NUMPY_TRANSLATIONS.
        if len(expr.args) == 1:
            return self._print(expr.args[0])
        result = 'logical_and({0}, {1})'.format(self._print(expr.args[-2]), self._print(expr.args[-1]))
        for arg in reversed(expr.args[:-2]):
            result = 'logical_and({0}, {1})'.format(self._print(arg), result)
        return  result

    def _print_Or(self, expr):
        "Logical Or printer"
        # We have to override LambdaPrinter because it uses Python 'or' keyword.
        # If LambdaPrinter didn't define it, we could use StrPrinter's
        # version of the function and add 'logical_or' to NUMPY_TRANSLATIONS.
        if len(expr.args) == 1:
            return self._print(expr.args[0])
        result = 'logical_or({0}, {1})'.format(self._print(expr.args[-2]), self._print(expr.args[-1]))
        for arg in reversed(expr.args[:-2]):
            result = 'logical_or({0}, {1})'.format(self._print(arg), result)
        return  result

    def _print_Not(self, expr):
        "Logical Not printer"
        # We have to override LambdaPrinter because it uses Python 'not' keyword.
        # If LambdaPrinter didn't define it, we would still have to define our
        #     own because StrPrinter doesn't define it.
        if len(expr.args) == 1:
            return self._print(expr.args[0])
        result = 'logical_not({0}, {1})'.format(self._print(expr.args[-2]), self._print(expr.args[-1]))
        for arg in reversed(expr.args[:-2]):
            result = 'logical_not({0}, {1})'.format(self._print(arg), result)
        return  result


def point_sample(comp_count, pdof=10):
    """
    Sample 'pdof * (sum(comp_count) - len(comp_count))' points in
    composition space for the sublattice configuration specified
    by 'comp_count'. Points are sampled quasi-randomly from a Halton sequence.
    A Halton sequence is like a uniform random distribution, but the
    result will always be the same for a given 'comp_count' and 'pdof'.
    Note: For systems with only one component, only one point will be
    returned, regardless of 'pdof'. This is because the degrees of freedom
    are zero for that case.

    Parameters
    ----------
    comp_count : list
        Number of components in each sublattice.
    pdof : int
        Number of points to sample per degree of freedom.

    Returns
    -------
    ndarray of generated points satisfying the mass balance.

    Examples
    --------
    >>> comps = [8,1] # 8 components in sublattice 1; only 1 in sublattice 2
    >>> pts = point_sample(comps, pdof=20) # 7 d.o.f, returns a 140x7 ndarray
    """
    # Generate Halton sequence with appropriate dimensions and size
    pts = halton(sum(comp_count),
                 pdof * (sum(comp_count) - len(comp_count)), scramble=True)
    # Convert low-discrepancy sequence to normalized exponential
    # This will be uniformly distributed over the simplices
    pts = -np.log(pts)
    cur_idx = 0
    for ctx in comp_count:
        end_idx = cur_idx + ctx
        pts[:, cur_idx:end_idx] /= pts[:, cur_idx:end_idx].sum(axis=1)[:, None]
        cur_idx = end_idx

    if len(pts) == 0:
        pts = np.atleast_2d([1] * len(comp_count))
    return pts

def make_callable(model, variables, mode=None):
    """
    Take a SymPy object and create a callable function.

    Parameters
    ----------
    model, SymPy object
        Abstract representation of function
    variables, list
        Input variables, ordered in the way the return function will expect
    mode, ['numpy', 'numba', 'sympy'], optional
        Method to use when 'compiling' the function. SymPy mode is
        slow and should only be used for debugging. If Numba is installed,
        it can offer speed-ups when calling the energy function many
        times on multi-core CPUs.

    Returns
    -------
    Function that takes arguments in the same order as 'variables'
    and returns the energy according to 'model'.

    Examples
    --------
    None yet.
    """
    energy = None
    if mode is None:
        # no mode specified; use numba if available, otherwise numpy
        if _NUMBA:
            mode = 'numba'
        else:
            mode = 'numpy'

    if mode == 'sympy':
        energy = lambda *vs: model.subs(zip(variables, vs)).evalf()
    elif mode == 'numpy':
        energy = lambdify(tuple(variables), model, dummify=True,
                          modules=[{'where': np.where, 'Abs': np.abs}, 'numpy'], printer=NumPyPrinter)
    elif mode == 'numba':
        variables = tuple(variables)
        varsig = 'float64({})'.format(','.join(['float64'] * len(variables)))
        energy = lambdify(variables, model, dummify=True,
                          modules=[{'where': nbwhere, 'Abs': np.abs}, 'numpy'], printer=NumPyPrinter)
        # target=parallel seems to incur too much overhead on small arrays
        energy = _NUMBA.vectorize([varsig], nopython=True, target='cpu')(energy)
    else:
        energy = lambdify(tuple(variables), model, dummify=True,
                          modules=mode)

    return energy

def sizeof_fmt(num, suffix='B'):
    """
    Human-readable string for a number of bytes.
    http://stackoverflow.com/questions/1094841/reusable-library-to-get-human-readable-version-of-file-size
    """
    for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
        if abs(num) < 1000.0:
            return "%3.1f%s%s" % (num, unit, suffix)
        num /= 1000.0
    return "%.1f%s%s" % (num, 'Y', suffix)

def unpack_condition(tup):
    """
    Convert a condition to a list of values.

    Notes
    -----
    Rules for keys of conditions dicts:
    (1) If it's numeric, treat as a point value
    (2) If it's a tuple with one element, treat as a point value
    (3) If it's a tuple with two elements, treat as lower/upper limits and guess a step size.
    (4) If it's a tuple with three elements, treat as lower/upper/step
    (5) If it's a list, ndarray or other non-tuple ordered iterable, use those values directly.

    """
    if isinstance(tup, tuple):
        if len(tup) == 1:
            return [float(tup[0])]
        elif len(tup) == 2:
            return np.arange(tup[0], tup[1], dtype=np.float)
        elif len(tup) == 3:
            return np.arange(tup[0], tup[1], tup[2], dtype=np.float)
        else:
            raise ValueError('Condition tuple is length {}'.format(len(tup)))
    elif isinstance(tup, collections.Iterable):
        return [float(x) for x in tup]
    else:
        return [float(tup)]

def unpack_phases(phases):
    "Convert a phases list/dict into a sorted list."
    active_phases = None
    if isinstance(phases, (list, tuple, set)):
        active_phases = sorted(phases)
    elif isinstance(phases, dict):
        active_phases = sorted([phn for phn, status in phases.items() \
            if status == 'entered'])
    elif type(phases) is str:
        active_phases = [phases]
    return active_phases

def generate_dof(phase, active_comps):
    """
    Accept a Phase object and a set() of the active components.
    Return a tuple of variable names and the sublattice degrees of freedom.
    """
    variables = []
    sublattice_dof = []
    for idx, sublattice in enumerate(phase.constituents):
        dof = 0
        for component in sorted(set(sublattice).intersection(active_comps)):
            variables.append(v.SiteFraction(phase.name.upper(), idx, component))
            dof += 1
        sublattice_dof.append(dof)
    return variables, sublattice_dof

def endmember_matrix(dof, vacancy_indices=None):
    """
    Accept the number of components in each sublattice.
    Return a matrix corresponding to the compositions of all endmembers.

    Parameters
    ----------
    dof : list of int
        Number of components in each sublattice.
    vacancy_indices, list of int, optional
        If vacancies are present in every sublattice, specify their indices
        in each sublattice to ensure the "pure vacancy" endmember is excluded.

    Examples
    --------
    Sublattice configuration like: `(AL, NI, VA):(AL, NI, VA):(VA)`
    >>> endmember_matrix([3,3,1], vacancy_indices=[2, 2, 0])
    """
    total_endmembers = functools.reduce(operator.mul, dof, 1)
    res_matrix = np.empty((total_endmembers, sum(dof)), dtype=np.float)
    dof_arrays = [np.eye(d).tolist() for d in dof]
    row_idx = 0
    for row in itertools.product(*dof_arrays):
        res_matrix[row_idx, :] = np.concatenate(row, axis=0)
        row_idx += 1
    if vacancy_indices is not None and len(vacancy_indices) == len(dof):
        dof_adj = np.array([sum(dof[0:i]) for i in range(len(dof))])
        indices = np.array(vacancy_indices) + dof_adj
        row_idx_to_delete = np.where(np.all(res_matrix[:, indices] == 1,
                                            axis=1))
        res_matrix = np.delete(res_matrix, (row_idx_to_delete), axis=0)
    # Adjust site fractions to the numerical limit
    cur_idx = 0
    res_matrix[res_matrix == 0] = MIN_SITE_FRACTION
    for ctx in dof:
        end_idx = cur_idx + ctx
        res_matrix[:, cur_idx:end_idx] /= res_matrix[:, cur_idx:end_idx].sum(axis=1)[:, None]
        cur_idx = end_idx
    return res_matrix

def unpack_kwarg(kwarg_obj, default_arg=None):
    """
    Keyword arguments in pycalphad can be passed as a constant value, a
    dict of phase names and values, or a list containing both of these. If
    the latter, then the dict is checked first; if the phase of interest is not
    there, then the constant value is used.

    This function is a way to construct defaultdicts out of keyword arguments.

    Parameters
    ----------
    kwarg_obj : dict, iterable, or None
        Argument to unpack into a defaultdict
    default_arg : object
        Default value to use if iterable isn't specified

    Returns
    -------
    defaultdict for the keyword argument of interest

    Examples
    --------
    >>> test_func = lambda **kwargs: print(unpack_kwarg('opt'))
    >>> test_func(opt=100)
    >>> test_func(opt={'FCC_A1': 50, 'BCC_B2': 10})
    >>> test_func(opt=[{'FCC_A1': 30}, 200])
    >>> test_func()
    >>> test_func2 = lambda **kwargs: print(unpack_kwarg('opt', default_arg=1))
    >>> test_func2()
    """
    new_dict = collections.defaultdict(lambda: default_arg)

    if isinstance(kwarg_obj, collections.Mapping):
        new_dict.update(kwarg_obj)
    # kwarg_obj is a list containing a dict and a default
    # For now at least, we don't treat ndarrays the same as other iterables
    # ndarrays are assumed to be numeric arrays containing "default values", so don't match here
    elif isinstance(kwarg_obj, collections.Iterable) and not isinstance(kwarg_obj, np.ndarray):
        for element in kwarg_obj:
            if isinstance(element, collections.Mapping):
                new_dict.update(element)
            else:
                # element=element syntax to silence var-from-loop warning
                new_dict = collections.defaultdict(
                    lambda element=element: element, new_dict)
    elif kwarg_obj is None:
        pass
    else:
        new_dict = collections.defaultdict(lambda: kwarg_obj)

    return new_dict




"""
This module defines valid Thermo-Calc TDB keywords and handles abbreviations.
Note that not all of these keywords are fully supported yet.
"""

import re

# Reference: Thermo-Calc Database Manager's Guide
TDB_KEYWORDS = sorted([
    'ELEMENT',
    'SPECIES',
    'PHASE',
    'CONSTITUENT',
    'ADD_CONSTITUENT',
    'COMPOUND_PHASE',
    'ALLOTROPIC_PHASE',
    'TEMPERATURE_LIMITS',
    'DEFINE_SYSTEM_DEFAULT',
    'DEFAULT_COMMAND',
    'DATABASE_INFORMATION',
    'TYPE_DEFINITION',
    'FTP_FILE',
    'FUNCTION',
    'PARAMETER',
    'OPTIONS',
    'TABLE',
    'ASSESSED_SYSTEMS',
    'REFERENCE_FILE',
    'LIST_OF_REFERENCES',
    'ADD_REFERENCE',
    'CASE',
    'ENDCASE',
    'VERSION_DATA',
    'VERSION_DATE',
    'DIFFUSION',
    'ZERO_VOLUME_SPECIES'
])

# Reference: Thermo-Calc Console Mode Command Reference, Version 4.1
TDB_COMMANDS = sorted([
    'AMEND_PHASE_DESCRIPTION'
    'DEFINE_ELEMENTS'
])

# Reference: Thermo-Calc Console Mode Command Reference, Version 4.1
TDB_PHASE_DESCRIPTIONS = sorted([
    'EXCESS_MODEL',
    'MAGNETIC_ORDERING',
    'DEBYE_HUCKEL',
    'STATUS_BITS',
    'NEW_CONSTITUENT',
    'RENAME_PHASE',
    'COMPOSITION_SETS',
    'GLASS_TRANSITION',
    'DISORDERED_PART',
    'MAJOR_CONSTITUENT',
    'ZRO2_TRANSITION',
    'REMOVE_ADDITION',
    'QUASICHEM_IONIC',
    'QUASICHEM_FACT00',
    'QUASICHEM_IRSID',
    'TERNARY_EXTRAPOLAT',
    'HKF_ELECTROSTATIC',
    'DEFAULT_STABLE',
    'SITE_RATIOS',
    'FRACTION_LIMITS'
])

TDB_PARAM_TYPES = sorted([
    'G',
    'L',
    'TC',
    'BMAGN',
    'MQ',
    'MF',
    'DQ',
    'DF',
    'V0',
    'VS'
])

def expand_keyword(possible, candidate):
    """
    Expand an abbreviated keyword based on the provided list.

    Parameters
    ----------
    possible : list of str
        Possible keywords for 'candidate' to be matched against.
    candidate : str
        Abbreviated keyword to expand.

    Returns
    -------
    list of str of matching expanded keywords

    Examples
    --------
    None yet.
    """
    # Rewritten to escape each token in the split string, instead of the input
    # The reason is that Python 2.7 will escape underscore characters
    candidate_pieces = [re.escape(x) for x in \
        candidate.upper().replace('-', '_').split('_')]
    pattern = r'^'
    pattern += r'[^_\s]*_'.join(candidate_pieces)
    pattern += r'[^\s]*$'
    matches = [re.match(pattern, pxd) for pxd in possible]
    matches = [m.string for m in matches if m is not None]
    if len(matches) == 0:
        raise ValueError('{0} does not match {1}'.format(candidate, possible))

    return matches



"""
This module constructs gradient functions for Models.
"""
from sympy import lambdify
from pycalphad.core.utils import NumPyPrinter
import numba
import numpy as np
import itertools

@numba.jit('float64(boolean, float64, float64)')
def where(condition, x, y):
    if condition:
        return x
    else:
        return y

def get_broadcast_shape(*shapes):
    """
    Given a set of array shapes, return the shape of the output when arrays of those
    shapes are broadcast together.
    Source: http://stackoverflow.com/questions/27196672/numpy-broadcast-indices-from-shape
    """
    max_nim = max(len(s) for s in shapes)
    equal_len_shapes = np.array([(1, )*(max_nim-len(s))+s for s in shapes])
    max_dim_shapes = np.max(equal_len_shapes, axis = 0)
    assert np.all(np.bitwise_or(equal_len_shapes==1, equal_len_shapes == max_dim_shapes[None, :])), \
        'Shapes %s are not broadcastable together' % (shapes, )
    return tuple(max_dim_shapes)

def make_gradient_from_graph(sympy_graph, variables):
    wrt = tuple(variables)
    grads = np.empty((len(wrt)), dtype=object)
    namespace = {}
    for i, mgrad in enumerate(sympy_graph.diff(vv) for vv in variables):
        grads[i] = mgrad
        gen_func = lambdify(tuple(wrt), grads[i], dummify=True, modules=[{'where': where,
                                                                          'Abs': np.abs}, 'numpy'], printer=NumPyPrinter)
        namespace['grad_{0}'.format(i)] = numba.jit(['float64({})'.format(','.join(['float64'] * len(wrt)))], nopython=True)\
            (gen_func)
    # Build the gradient using compile() and exec
    # We do this because Numba needs "static" information about the arguments and functions
    call_args = ','.join(['_x{0}'.format(i) for i in range(len(wrt))])
    call_passed_args = ','.join(['_x{0}[0]'.format(i) for i in range(len(wrt))])

    grad_code = 'def grad_func({0}, lengthfix, result):'.format(call_args)
    grad_list = ['    result[{0}] = grad_{0}({1})'.format(i, call_passed_args) for i in range(len(wrt))]
    grad_code = grad_code + '\n' + '\n'.join(grad_list)

    grad_code = compile(grad_code, '<string>', 'exec')
    exec(grad_code, namespace)
    grad_func = numba.guvectorize([','.join(['float64[:]'] * (len(wrt)+2))],
                                   ','.join(['()'] * len(wrt)) + ',(n)->(n)', nopython=True)(namespace['grad_func'])
    hess_code = 'def hess_func({0}, lengthfix, result):'.format(call_args)
    hess_list = []
    for i in range(len(wrt)):
        for j in range(i,len(wrt)):
            # We have to do a little bit of trickery here to make sure we don't exceed the domain of our variables
            finite_diff_args = ['_x{0}[0]'.format(g) for g in range(len(wrt))]
            finite_diff_args[j] = '_x{0}[0]+1e-14'.format(j)
            finite_diff_args = ','.join(finite_diff_args)
            hess_list.append('    result[{0},{1}] = result[{1},{0}] = ((grad_{0}({2}) - grad_{0}({3}))/1e-14)'
                             .format(i, j, finite_diff_args, call_passed_args))
    hess_code = hess_code + '\n' + '\n'.join(hess_list)
    hess_code = compile(hess_code, '<string>', 'exec')
    namespace['where'] = where
    exec(hess_code, namespace)
    hess_func = numba.guvectorize([','.join(['float64[:]'] * (len(wrt) + 1)) + ',float64[:,:]'],
                                  ','.join(['()'] * len(wrt)) + ',(n)->(n,n)', nopython=True)(namespace['hess_func'])
    def grad_argwrapper(*args, out=None):
        result_shape = get_broadcast_shape(*[np.asarray(arg).shape for arg in args])
        result_shape = result_shape + (len(args),)
        result = out if out is not None else np.zeros(result_shape, dtype=np.float)
        return grad_func(*list(itertools.chain(args, [np.empty(len(args)), result])))
    def hess_argwrapper(*args, out=None):
        result_shape = get_broadcast_shape(*[np.asarray(arg).shape for arg in args])
        result_shape = result_shape + (len(args),len(args))
        result = out if out is not None else np.zeros(result_shape, dtype=np.float)
        hess_func(*list(itertools.chain(args, [np.empty(len(args)), result])))
        return result
    return grad_argwrapper, hess_argwrapper


"""
The constants module contains some numerical constants for use
in the module.
Note that modifying these may yield unpredictable results.
"""
# Force zero values to this amount, for numerical stability
MIN_SITE_FRACTION = 1e-12
# For each phase pair with composition difference below tolerance, eliminate phase with largest index
COMP_DIFFERENCE_TOL = 0.01


"""
The Model module provides support for using a Database to perform
calculations under specified conditions.
"""
from __future__ import division
import copy
from sympy import log, Abs, Add, Mul, Piecewise, Pow, S, sin, Symbol
from tinydb import where
import pycalphad.variables as v
from pycalphad.core.constants import MIN_SITE_FRACTION
from pycalphad.log import logger
import numpy as np

# Maximum number of levels deep we check for symbols that are functions of
# other symbols
_MAX_PARAM_NESTING = 32

class DofError(Exception):
    "Error due to missing degrees of freedom."
    pass

class Model(object):
    """
    Models use an abstract representation of the function
    for calculation of values under specified conditions.

    Parameters
    ----------
    dbf : Database
        Database containing the relevant parameters.
    comps : list
        Names of components to consider in the calculation.
    phase : list
        Name of phase model to build.
    parameters : dict
        Optional dictionary of parameters to be substituted in the model.
        This will overwrite parameters specified in the database

    Methods
    -------
    None yet.

    Examples
    --------
    None yet.
    """
    def __init__(self, dbe, comps, phase, parameters=None):
        # Constrain possible components to those within phase's d.o.f
        possible_comps = set([x.upper() for x in comps])
        self.components = set()
        self.constituents = []
        self.phase_name = phase.upper()
        self.site_ratios = dbe.phases[phase.upper()].sublattices
        for sublattice in dbe.phases[phase.upper()].constituents:
            self.components |= set(sublattice).intersection(possible_comps)
        logger.debug('Model of %s has components %s', phase, self.components)
        # Verify that this phase is still possible to build
        for sublattice in dbe.phases[phase.upper()].constituents:
            if len(set(sublattice).intersection(self.components)) == 0:
                # None of the components in a sublattice are active
                # We cannot build a model of this phase
                raise DofError(
                    '{0}: Sublattice {1} of {2} has no components in {3}' \
                    .format(phase.upper(), sublattice,
                            dbe.phases[phase.upper()].constituents,
                            self.components))
            self.constituents.append(set(sublattice).intersection(self.components))

        # Convert string symbol names to sympy Symbol objects
        # This makes xreplace work with the symbols dict
        symbols = dict([(Symbol(s), val) for s, val in dbe.symbols.items()])

        if parameters is not None:
            symbols.update([(Symbol(s), val) for s, val in parameters.items()])

        self.models = dict()
        self.build_phase(dbe, phase.upper(), symbols, dbe.search)
        self.site_fractions = sorted(self.ast.atoms(v.SiteFraction), key=str)

        # Need to do more substitutions to catch symbols that are functions
        # of other symbols
        for name, value in self.models.items():
            try:
                for iteration in range(_MAX_PARAM_NESTING):
                    self.models[name] = self.models[name].xreplace(symbols)
                    undefs = self.models[name].atoms(Symbol) - self.models[name].atoms(v.StateVariable)
                    if len(undefs) == 0:
                        break
            except AttributeError:
                # Can't use xreplace on a float
                pass

    def __eq__(self, other):
        if self is other:
            return True
        elif type(self) != type(other):
            return False
        else:
            return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(repr(self))

    def standard_mole_fraction(self, species):
        "Mole fraction which correctly normalizes for vacancies."
        result = S.Zero
        site_ratio_normalization = S.Zero
        # Calculate normalization factor
        for idx, sublattice in enumerate(self.constituents):
            active = set(sublattice).intersection(self.components)
            if 'VA' in active:
                site_ratio_normalization += self.site_ratios[idx] * \
                    (1.0 - v.SiteFraction(self.phase_name, idx, 'VA'))
            else:
                site_ratio_normalization += self.site_ratios[idx]
        site_ratios = [c/site_ratio_normalization for c in self.site_ratios]
        # Sum up site fraction contributions from each component sublattice
        for idx, sublattice in enumerate(self.constituents):
            active = set(sublattice).intersection(set(self.components))
            if species in active:
                result += site_ratios[idx] * \
                    v.SiteFraction(self.phase_name, idx, species)
        return result

    @property
    def ast(self):
        "Return the full abstract syntax tree of the model."
        return Add(*list(self.models.values()))

    @property
    def variables(self):
        "Return state variables in the model."
        return sorted(self.ast.atoms(v.StateVariable), key=str)

    @property
    def degree_of_ordering(self):
        result = S.Zero
        site_ratio_normalization = S.Zero
        # Calculate normalization factor
        for idx, sublattice in enumerate(self.constituents):
            active = set(sublattice).intersection(self.components)
            if 'VA' in active:
                site_ratio_normalization += self.site_ratios[idx] * \
                    (1.0 - v.SiteFraction(self.phase_name, idx, 'VA'))
            else:
                site_ratio_normalization += self.site_ratios[idx]
        site_ratios = [c/site_ratio_normalization for c in self.site_ratios]
        for comp in sorted([c for c in self.components if c != 'VA']):
            comp_result = S.Zero
            for idx, sublattice in enumerate(self.constituents):
                active = set(sublattice).intersection(set(self.components))
                if comp in active:
                    comp_result += site_ratios[idx] * Abs(v.SiteFraction(self.phase_name, idx, comp) - self.standard_mole_fraction(comp)) / self.standard_mole_fraction(comp)
            result += comp_result
        return result / len([c for c in self.components if c != 'VA'])
    DOO = degree_of_ordering

    # Can be defined as a list of pre-computed first derivatives
    gradient = None

    # Note: In order-disorder phases, TC will always be the *disordered* value of TC
    curie_temperature = TC = S.Zero

    #pylint: disable=C0103
    # These are standard abbreviations from Thermo-Calc for these quantities
    energy = GM = property(lambda self: self.ast)
    entropy = SM = property(lambda self: -self.GM.diff(v.T))
    enthalpy = HM = property(lambda self: self.GM - v.T*self.GM.diff(v.T))
    heat_capacity = CPM = property(lambda self: -v.T*self.GM.diff(v.T, v.T))
    #pylint: enable=C0103
    mixing_energy = GM_MIX = property(lambda self: self.GM - self.models['ref'])
    mixing_enthalpy = HM_MIX = \
        property(lambda self: self.GM_MIX - v.T*self.GM_MIX.diff(v.T))
    mixing_entropy = SM_MIX = property(lambda self: -self.GM_MIX.diff(v.T))
    mixing_heat_capacity = CPM_MIX = \
        property(lambda self: -v.T*self.GM_MIX.diff(v.T, v.T))

    def build_phase(self, dbe, phase_name, symbols, param_search):
        """
        Apply phase's model hints to build a master SymPy object.
        """
        phase = dbe.phases[phase_name]
        self.models['ref'] = self.reference_energy(phase, param_search)
        self.models['idmix'] = self.ideal_mixing_energy(phase, param_search)
        self.models['xsmix'] = self.excess_mixing_energy(phase, param_search)
        self.models['mag'] = self.magnetic_energy(phase, param_search)

        # Next, we handle atomic ordering
        # NOTE: We need to add this one last since it uses self.models
        # to figure out the contribution.
        # It will also MODIFY self.models
        ordered_phase_name = None
        disordered_phase_name = None
        try:
            ordered_phase_name = phase.model_hints['ordered_phase']
            disordered_phase_name = phase.model_hints['disordered_phase']
        except KeyError:
            pass
        if ordered_phase_name == phase_name:
            self.models['ord'] = self.atomic_ordering_energy(dbe,
                                                             disordered_phase_name,
                                                             ordered_phase_name)

    def _purity_test(self, constituent_array):
        """
        Check if constituent array only has one species in its array
        This species must also be an active species
        """
        for sublattice in constituent_array:
            if len(sublattice) != 1:
                return False
            if (sublattice[0] not in self.components) and \
                (sublattice[0] != '*'):
                return False
        return True

    def _array_validity(self, constituent_array):
        """
        Check that the current array contains only active species.
        """
        for sublattice in constituent_array:
            valid = set(sublattice).issubset(self.components) \
                or sublattice[0] == '*'
            if not valid:
                return False
        return True

    def _interaction_test(self, constituent_array):
        """
        Check if constituent array has more than one active species in
        its array for at least one sublattice.
        """
        result = False
        for sublattice in constituent_array:
            # check if all elements involved are also active
            valid = set(sublattice).issubset(self.components) \
                or sublattice[0] == '*'
            if len(sublattice) > 1 and valid:
                result = True
            if not valid:
                result = False
                break
        return result

    def _site_ratio_normalization(self, phase):
        """
        Calculates the normalization factor based on the number of sites
        in each sublattice.
        """
        site_ratio_normalization = S.Zero
        # Normalize by the sum of site ratios times a factor
        # related to the site fraction of vacancies
        for idx, sublattice in enumerate(phase.constituents):
            if ('VA' in set(sublattice)) and ('VA' in self.components):
                site_ratio_normalization += phase.sublattices[idx] * \
                    (1 - v.SiteFraction(phase.name, idx, 'VA'))
            else:
                site_ratio_normalization += phase.sublattices[idx]
        return site_ratio_normalization

    @staticmethod
    def _Muggianu_correction_dict(comps): #pylint: disable=C0103
        """
        Replace y_i -> y_i + (1 - sum(y involved in parameter)) / m,
        where m is the arity of the interaction parameter.
        Returns a dict converting the list of Symbols (comps) to this.
        m is assumed equal to the length of comps.

        When incorporating binary, ternary or n-ary interaction parameters
        into systems with more than n components, the sum of site fractions
        involved in the interaction parameter may no longer be unity. This
        breaks the symmetry of the parameter. The solution suggested by
        Muggianu, 1975, is to renormalize the site fractions by replacing them
        with a term that will sum to unity even in higher-order systems.
        There are other solutions that involve retaining the asymmetry for
        physical reasons, but this solution works well for components that
        are physically similar.

        This procedure is based on an analysis by Hillert, 1980,
        published in the Calphad journal.
        """
        arity = len(comps)
        return_dict = {}
        correction_term = (S.One - Add(*comps)) / arity
        for comp in comps:
            return_dict[comp] = comp + correction_term
        return return_dict

    def redlich_kister_sum(self, phase, param_search, param_query):
        """
        Construct parameter in Redlich-Kister polynomial basis, using
        the Muggianu ternary parameter extension.
        """
        rk_terms = []

        # search for desired parameters
        params = param_search(param_query)
        for param in params:
            # iterate over every sublattice
            mixing_term = S.One
            for subl_index, comps in enumerate(param['constituent_array']):
                comp_symbols = None
                # convert strings to symbols
                if comps[0] == '*':
                    # Handle wildcards in constituent array
                    comp_symbols = \
                        [
                            v.SiteFraction(phase.name, subl_index, comp)
                            for comp in set(phase.constituents[subl_index])\
                                .intersection(self.components)
                        ]
                    mixing_term *= Add(*comp_symbols)
                else:
                    comp_symbols = \
                        [
                            v.SiteFraction(phase.name, subl_index, comp)
                            for comp in comps
                        ]
                    mixing_term *= Mul(*comp_symbols)
                # is this a higher-order interaction parameter?
                if len(comps) == 2 and param['parameter_order'] > 0:
                    # interacting sublattice, add the interaction polynomial
                    mixing_term *= Pow(comp_symbols[0] - \
                        comp_symbols[1], param['parameter_order'])
                if len(comps) == 3:
                    # 'parameter_order' is an index to a variable when
                    # we are in the ternary interaction parameter case

                    # NOTE: The commercial software packages seem to have
                    # a "feature" where, if only the zeroth
                    # parameter_order term of a ternary parameter is specified,
                    # the other two terms are automatically generated in order
                    # to make the parameter symmetric.
                    # In other words, specifying only this parameter:
                    # PARAMETER G(FCC_A1,AL,CR,NI;0) 298.15  +30300; 6000 N !
                    # Actually implies:
                    # PARAMETER G(FCC_A1,AL,CR,NI;0) 298.15  +30300; 6000 N !
                    # PARAMETER G(FCC_A1,AL,CR,NI;1) 298.15  +30300; 6000 N !
                    # PARAMETER G(FCC_A1,AL,CR,NI;2) 298.15  +30300; 6000 N !
                    #
                    # If either 1 or 2 is specified, no implicit parameters are
                    # generated.
                    # We need to handle this case.
                    if param['parameter_order'] == 0:
                        # are _any_ of the other parameter_orders specified?
                        ternary_param_query = (
                            (where('phase_name') == param['phase_name']) & \
                            (where('parameter_type') == \
                                param['parameter_type']) & \
                            (where('constituent_array') == \
                                param['constituent_array'])
                        )
                        other_tern_params = param_search(ternary_param_query)
                        if len(other_tern_params) == 1 and \
                            other_tern_params[0] == param:
                            # only the current parameter is specified
                            # We need to generate the other two parameters.
                            order_one = copy.deepcopy(param)
                            order_one['parameter_order'] = 1
                            order_two = copy.deepcopy(param)
                            order_two['parameter_order'] = 2
                            # Add these parameters to our iteration.
                            params.extend((order_one, order_two))
                    # Include variable indicated by parameter order index
                    # Perform Muggianu adjustment to site fractions
                    mixing_term *= comp_symbols[param['parameter_order']].subs(
                        self._Muggianu_correction_dict(comp_symbols),
                        simultaneous=True)
            rk_terms.append(mixing_term * param['parameter'])
        return Add(*rk_terms)

    def reference_energy(self, phase, param_search):
        """
        Returns the weighted average of the endmember energies
        in symbolic form.
        """
        pure_param_query = (
            (where('phase_name') == phase.name) & \
            (where('parameter_order') == 0) & \
            (where('parameter_type') == "G") & \
            (where('constituent_array').test(self._purity_test))
        )
        pure_energy_term = self.redlich_kister_sum(phase, param_search,
                                                   pure_param_query)
        return pure_energy_term / self._site_ratio_normalization(phase)

    def ideal_mixing_energy(self, phase, param_search):
        #pylint: disable=W0613
        """
        Returns the ideal mixing energy in symbolic form.
        """
        # Normalize site ratios
        site_ratio_normalization = self._site_ratio_normalization(phase)
        site_ratios = phase.sublattices
        site_ratios = [c/site_ratio_normalization for c in site_ratios]
        ideal_mixing_term = S.Zero
        for subl_index, sublattice in enumerate(phase.constituents):
            active_comps = set(sublattice).intersection(self.components)
            if len(active_comps) == 1:
                continue # no mixing if only one species in sublattice
            ratio = site_ratios[subl_index]
            for comp in active_comps:
                sitefrac = \
                    v.SiteFraction(phase.name, subl_index, comp)
                # We lose some precision here, but this makes the limit behave nicely
                # We're okay until fractions of about 1e-12 (platform-dependent)
                mixing_term = Piecewise((sitefrac*log(sitefrac), sitefrac > MIN_SITE_FRACTION/10.), (0, True))
                ideal_mixing_term += (mixing_term*ratio)
        ideal_mixing_term *= (v.R * v.T)
        return ideal_mixing_term

    def excess_mixing_energy(self, phase, param_search):
        """
        Build the binary, ternary and higher order interaction term
        Here we use Redlich-Kister polynomial basis by default
        Here we use the Muggianu ternary extension by default
        Replace y_i -> y_i + (1 - sum(y involved in parameter)) / m,
        where m is the arity of the interaction parameter
        """
        param_query = (
            (where('phase_name') == phase.name) & \
                ((where('parameter_type') == 'G') |
                 (where('parameter_type') == 'L')) & \
                (where('constituent_array').test(self._interaction_test))
            )
        excess_term = self.redlich_kister_sum(phase, param_search, param_query)
        return excess_term / self._site_ratio_normalization(phase)

    def magnetic_energy(self, phase, param_search):
        #pylint: disable=C0103, R0914
        """
        Return the energy from magnetic ordering in symbolic form.
        The implemented model is the Inden-Hillert-Jarl formulation.
        The approach follows from the background section of W. Xiong, 2011.
        """
        self.TC = self.curie_temperature = S.Zero
        if 'ihj_magnetic_structure_factor' not in phase.model_hints:
            return S.Zero
        if 'ihj_magnetic_afm_factor' not in phase.model_hints:
            return S.Zero

        site_ratio_normalization = self._site_ratio_normalization(phase)
        # define basic variables
        afm_factor = phase.model_hints['ihj_magnetic_afm_factor']

        bm_param_query = (
            (where('phase_name') == phase.name) & \
            (where('parameter_type') == 'BMAGN') & \
            (where('constituent_array').test(self._array_validity))
        )
        tc_param_query = (
            (where('phase_name') == phase.name) & \
            (where('parameter_type') == 'TC') & \
            (where('constituent_array').test(self._array_validity))
        )

        mean_magnetic_moment = \
            self.redlich_kister_sum(phase, param_search, bm_param_query)
        beta = mean_magnetic_moment / Piecewise(
            (afm_factor, mean_magnetic_moment <= 0),
            (1., True)
            )

        curie_temp = \
            self.redlich_kister_sum(phase, param_search, tc_param_query)
        tc = curie_temp / Piecewise(
            (afm_factor, curie_temp <= 0),
            (1., True)
            )
        self.TC = self.curie_temperature = tc
        #print(tc)
        # 0.01 used to prevent singularity
        tau = v.T / (tc + 0.01)

        # define model parameters
        p = phase.model_hints['ihj_magnetic_structure_factor']
        A = 518/1125 + (11692/15975)*(1/p - 1)
        # factor when tau < 1
        sub_tau = 1 - (1/A) * ((79/(140*p))*(tau**(-1)) + (474/497)*(1/p - 1) \
            * ((tau**3)/6 + (tau**9)/135 + (tau**15)/600)
                              )
        # factor when tau >= 1
        super_tau = -(1/A) * ((tau**-5)/10 + (tau**-15)/315 + (tau**-25)/1500)

        expr_cond_pairs = [(sub_tau, tau < 1),
                           (super_tau, True)
                           ]
        g_term = Piecewise(*expr_cond_pairs)

        return v.R * v.T * log(beta+1) * \
            g_term / site_ratio_normalization

    @staticmethod
    def mole_fraction(species_name, phase_name, constituent_array,
                      site_ratios):
        """
        Return an abstract syntax tree of the mole fraction of the
        given species as a function of its constituent site fractions.

        Note that this will treat vacancies the same as any other component,
        i.e., this will not give the correct _overall_ composition for
        sublattices containing vacancies with other components by normalizing
        by a factor of 1 - y_{VA}. This is because we use this routine in the
        order-disorder model to calculate the disordered site fractions from
        the ordered site fractions, so we need _all_ site fractions, including
        VA, to sum to unity.
        """

        # Normalize site ratios
        site_ratio_normalization = 0
        numerator = S.Zero
        for idx, sublattice in enumerate(constituent_array):
            # sublattices with only vacancies don't count
            if len(sublattice) == 1 and sublattice[0] == 'VA':
                continue
            if species_name in list(sublattice):
                site_ratio_normalization += site_ratios[idx]
                numerator += site_ratios[idx] * \
                    v.SiteFraction(phase_name, idx, species_name)

        if site_ratio_normalization == 0 and species_name == 'VA':
            return 1

        if site_ratio_normalization == 0:
            raise ValueError('Couldn\'t find ' + species_name + ' in ' + \
                str(constituent_array))

        return numerator / site_ratio_normalization

    def atomic_ordering_energy(self, dbe, disordered_phase_name,
                               ordered_phase_name):
        """
        Return the atomic ordering contribution in symbolic form.
        Description follows Servant and Ansara, Calphad, 2001.
        """
        disordered_model = self.__class__(dbe, self.components,
                                          disordered_phase_name)
        constituents = [sorted(set(c).intersection(self.components)) \
                for c in dbe.phases[ordered_phase_name].constituents]

        # Fix variable names
        variable_rename_dict = {}
        for atom in disordered_model.energy.atoms(v.SiteFraction):
            # Replace disordered phase site fractions with mole fractions of
            # ordered phase site fractions.
            # Special case: Pure vacancy sublattices
            all_species_in_sublattice = \
                dbe.phases[disordered_phase_name].constituents[
                    atom.sublattice_index]
            if atom.species == 'VA' and len(all_species_in_sublattice) == 1:
                # Assume: Pure vacancy sublattices are always last
                vacancy_subl_index = \
                    len(dbe.phases[ordered_phase_name].constituents)-1
                variable_rename_dict[atom] = \
                    v.SiteFraction(
                        ordered_phase_name, vacancy_subl_index, atom.species)
            else:
                # All other cases: replace site fraction with mole fraction
                variable_rename_dict[atom] = \
                    self.mole_fraction(
                        atom.species,
                        ordered_phase_name,
                        constituents,
                        dbe.phases[ordered_phase_name].sublattices
                        )
        # Save all of the ordered energy contributions
        # This step is why this routine must be called _last_ in build_phase
        ordered_energy = Add(*list(self.models.values()))
        self.models.clear()
        # Copy the disordered energy contributions into the correct bins
        for name, value in disordered_model.models.items():
            self.models[name] = value.xreplace(variable_rename_dict)
        # All magnetic parameters will be defined in the disordered model
        self.TC = self.curie_temperature = disordered_model.TC
        self.TC = self.curie_temperature = self.TC.xreplace(variable_rename_dict)

        molefraction_dict = {}

        # Construct a dictionary that replaces every site fraction with its
        # corresponding mole fraction in the disordered state
        for sitefrac in ordered_energy.atoms(v.SiteFraction):
            all_species_in_sublattice = \
                dbe.phases[ordered_phase_name].constituents[
                    sitefrac.sublattice_index]
            if sitefrac.species == 'VA' and len(all_species_in_sublattice) == 1:
                # pure-vacancy sublattices should not be replaced
                # this handles cases like AL,NI,VA:AL,NI,VA:VA and
                # ensures the VA's don't get mixed up
                continue
            molefraction_dict[sitefrac] = \
                self.mole_fraction(sitefrac.species,
                                   ordered_phase_name, constituents,
                                   dbe.phases[ordered_phase_name].sublattices)

        return ordered_energy - ordered_energy.subs(molefraction_dict,
                                                    simultaneous=True)

class TestModel(Model):
    """
    Test Model object for global minimization.

    Equation 15.2 in:
    P.M. Pardalos, H.E. Romeijn (Eds.), Handbook of Global Optimization,
    vol. 2. Kluwer Academic Publishers, Boston/Dordrecht/London (2002)

    Parameters
    ----------
    dbf : Database
        Ignored by TestModel but retained for API compatibility.
    comps : sequence
        Names of components to consider in the calculation.
    phase : str
        Name of phase model to build.
    solution : sequence, optional
        Float array locating the true minimum. Same length as 'comps'.
        If not specified, randomly generated and saved to self.solution

    Methods
    -------
    None yet.

    Examples
    --------
    None yet.
    """
    def __init__(self, dbf, comps, phase, solution=None, kmax=None):
        self.components = set([x.upper() for x in comps])
        if 'VA' in self.components:
            raise ValueError('Vacancies are unsupported in TestModel')
        self.models = dict()
        variables = [v.SiteFraction(phase.upper(), 0, x) for x in sorted(self.components)]
        if solution is None:
            solution = np.random.dirichlet(np.ones_like(variables, dtype=np.int))
        self.solution = dict(list(zip(variables, solution)))
        kmax = kmax if kmax is not None else 2
        scale_factor = 1e4 * len(self.components)
        ampl_scale = 1e3 * np.ones(kmax, dtype=np.float)
        freq_scale = 10 * np.ones(kmax, dtype=np.float)
        polys = Add(*[ampl_scale[i] * sin(freq_scale[i] * Add(*[Add(*[(varname - sol)**(j+1)
                                                                      for varname, sol in self.solution.items()])
                                                                for j in range(kmax)]))**2
                      for i in range(kmax)])
        self.models['test'] = scale_factor * Add(*[(varname - sol)**2 for varname, sol in self.solution.items()]) + polys




#pylint: disable=C0103,R0903,W0232
"""
Classes and constants for representing thermodynamic variables.
"""

from sympy import Float, Symbol

class StateVariable(Symbol):
    """
    State variables are symbols with built-in assumptions of being real.
    """
    def __new__(cls, name, **assumptions):
        return Symbol.__new__(cls, name.upper(), real=True, **assumptions)

class SiteFraction(StateVariable):
    """
    Site fractions are symbols with built-in assumptions of being real
    and nonnegative. The constructor handles formatting of the name.
    """
    def __new__(cls, phase_name, subl_index, species): #pylint: disable=W0221
        varname = phase_name + str(subl_index) + species
        #pylint: disable=E1121
        new_self = StateVariable.__new__(cls, varname, nonnegative=True)
        new_self.phase_name = phase_name.upper()
        new_self.sublattice_index = subl_index
        new_self.species = species.upper()
        return new_self
    def _latex(self):
        "LaTeX representation."
        #pylint: disable=E1101
        return 'y^{'+self.phase_name.replace('_', '-') + \
            '}_{'+str(self.subl_index)+'},_{'+self.species+'}'
    def __str__(self):
        "String representation."
        #pylint: disable=E1101
        return 'Y(%s,%d,%s)' % \
            (self.phase_name, self.sublattice_index, self.species)

class PhaseFraction(StateVariable):
    """
    Phase fractions are symbols with built-in assumptions of being real
    and nonnegative. The constructor handles formatting of the name.
    """
    def __new__(cls, phase_name, multiplicity): #pylint: disable=W0221
        varname = phase_name + str(multiplicity)
        #pylint: disable=E1121
        new_self = StateVariable.__new__(cls, varname, nonnegative=True)
        new_self.phase_name = phase_name.upper()
        new_self.multiplicity = multiplicity
        return new_self
    def _latex(self):
        "LaTeX representation."
        #pylint: disable=E1101
        return 'f^{'+self.phase_name.replace('_', '-') + \
            '}_{'+str(self.multiplicity)+'}'

class Composition(StateVariable):
    """
    Compositions are symbols with built-in assumptions of being real
    and nonnegative.
    """
    def __new__(cls, *args): #pylint: disable=W0221
        new_self = None
        varname = None
        phase_name = None
        species = None
        if len(args) == 1:
            # this is an overall composition variable
            species = args[0].upper()
            varname = 'X_' + species
        elif len(args) == 2:
            # this is a phase-specific composition variable
            phase_name = args[0].upper()
            species = args[1].upper()
            varname = 'X_' + phase_name + '_' + species
        else:
            # not defined
            raise ValueError('Composition not defined for args: '+args)

        #pylint: disable=E1121
        new_self = StateVariable.__new__(cls, varname, nonnegative=True)
        new_self.phase_name = phase_name
        new_self.species = species
        return new_self
    def _latex(self):
        "LaTeX representation."
        #pylint: disable=E1101
        if self.phase_name:
            return 'x^{'+self.phase_name.replace('_', '-') + \
                '}_{'+self.species+'}'
        else:
            return 'x_{'+self.species+'}'

class ChemicalPotential(StateVariable):
    """
    Chemical potentials are symbols with built-in assumptions of being real.
    """
    def __new__(cls, species, **assumptions):
        varname = 'MU_' + species.upper()
        new_self = StateVariable.__new__(cls, varname, **assumptions)
        new_self.species = species
        return new_self
    def _latex(self):
        "LaTeX representation."
        return '\mu_{'+self.species+'}'
    def __str__(self):
        "String representation."
        return 'MU(%s)' % self.species

temperature = T = StateVariable('T')
entropy = S = StateVariable('S')
pressure = P = StateVariable('P')
volume = V = StateVariable('V')
moles = N = StateVariable('N')
site_fraction = Y = SiteFraction
X = Composition
MU = ChemicalPotential
si_gas_constant = R = Float(8.3145) # ideal gas constant



from setuptools import setup
import os
import versioneer
versioneer.VCS = 'git'
versioneer.versionfile_source = 'pycalphad/_version.py'
versioneer.versionfile_build = 'pycalphad/_version.py'
versioneer.tag_prefix = '' # tags are like 1.2.0
versioneer.parentdir_prefix = 'pycalphad-' # dirname like 'myproject-1.2.0'

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name='pycalphad',
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author='Richard Otis',
    author_email='richard.otis@outlook.com',
    description='CALPHAD tools for designing thermodynamic models, calculating phase diagrams and investigating phase equilibria.',
    packages=['pycalphad', 'pycalphad.core', 'pycalphad.io', 'pycalphad.plot'],
    license='MIT',
    long_description=read('README.rst'),
    url='https://github.com/richardotis/pycalphad',
    install_requires=['matplotlib', 'pandas', 'xarray', 'sympy', 'pyparsing', 'tqdm',
                      'autograd', 'tinydb', 'scipy', 'numpy>=1.9'],
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Scientific/Engineering :: Chemistry',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5'
    ],

)
